# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

_ANALYSIS_PROTOCOL_TEXT = """<JaxbAnalysisProtocol>
    <Name>unnamed</Name>
    <JaxbAnalysisSettings>
        <Type>com.apldbio.sds.platform.core.analysis.IDetectorSettings</Type>
        <JaxbSettingValue>
            <Name>ConfidenceLevel</Name>
            <JaxbValueItem type="Double">
                <DoubleValue>99.0</DoubleValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>CreatedByUser</Name>
            <JaxbValueItem type="Boolean">
                <BooleanValue>true</BooleanValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>Threshold</Name>
            <JaxbValueItem type="Double">
                <DoubleValue>0.2</DoubleValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>AutoBaseline</Name>
            <JaxbValueItem type="Boolean">
                <BooleanValue>true</BooleanValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>BaselineStart</Name>
            <JaxbValueItem type="Integer">
                <IntValue>3</IntValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>ObjectName</Name>
            <JaxbValueItem type="String">
                <StringValue>AnalysisProtocol.DEFAULT_SETTINGS</StringValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>BaselineStop</Name>
            <JaxbValueItem type="Integer">
                <IntValue>15</IntValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>AutoCt</Name>
            <JaxbValueItem type="Boolean">
                <BooleanValue>true</BooleanValue>
            </JaxbValueItem>
        </JaxbSettingValue>
    </JaxbAnalysisSettings>
    <JaxbAnalysisSettings>
        <Type>com.apldbio.sds.platform.core.analysis.IWellSettings</Type>
        <JaxbSettingValue>
            <Name>BaselineStart</Name>
            <JaxbValueItem type="Integer">
                <IntValue>3</IntValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>ObjectName</Name>
            <JaxbValueItem type="String">
                <StringValue>AnalysisProtocol.DEFAULT_SETTINGS</StringValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>BaselineStop</Name>
            <JaxbValueItem type="Integer">
                <IntValue>15</IntValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>UseDetectorDefaults</Name>
            <JaxbValueItem type="Boolean">
                <BooleanValue>false</BooleanValue>
            </JaxbValueItem>
        </JaxbSettingValue>
    </JaxbAnalysisSettings>
    <JaxbAnalysisSettings>
        <Type>com.apldbio.sds.platform.core.analysis.ISignalSmoothingSettings</Type>
        <JaxbSettingValue>
            <Name>SignalSmoothing</Name>
            <JaxbValueItem type="Boolean">
                <BooleanValue>true</BooleanValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>ObjectName</Name>
            <JaxbValueItem type="String">
                <StringValue>AnalysisProtocol.DEFAULT_SETTINGS</StringValue>
            </JaxbValueItem>
        </JaxbSettingValue>
    </JaxbAnalysisSettings>
    <JaxbAnalysisSettings>
        <Type>com.apldbio.sds.platform.core.analysis.IAlgorithmSelectSettings</Type>
        <JaxbSettingValue>
            <Name>AlgorithmName</Name>
            <JaxbValueItem type="String">
                <StringValue>default</StringValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>ObjectName</Name>
            <JaxbValueItem type="String">
                <StringValue>AnalysisProtocol.DEFAULT_SETTINGS</StringValue>
            </JaxbValueItem>
        </JaxbSettingValue>
    </JaxbAnalysisSettings>
    <JaxbAnalysisSettings>
        <Type>com.apldbio.sds.platform.core.analysis.IAutoAnalysisSettings</Type>
        <JaxbSettingValue>
            <Name>AutoAnalysis</Name>
            <JaxbValueItem type="Boolean">
                <BooleanValue>false</BooleanValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>AutoSpectral</Name>
            <JaxbValueItem type="Boolean">
                <BooleanValue>false</BooleanValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>ObjectName</Name>
            <JaxbValueItem type="String">
                <StringValue>AnalysisProtocol.DEFAULT_SETTINGS</StringValue>
            </JaxbValueItem>
        </JaxbSettingValue>
    </JaxbAnalysisSettings>
    <JaxbAnalysisSettings>
        <Type>com.apldbio.sds.platform.core.analysis.IDataSelectSettings</Type>
        <JaxbSettingValue>
            <Name>StepNum</Name>
            <JaxbValueItem type="Integer">
                <IntValue>5</IntValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>StageNum</Name>
            <JaxbValueItem type="Integer">
                <IntValue>4</IntValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>MeltStageNum</Name>
            <JaxbValueItem type="Integer">
                <IntValue>0</IntValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>ObjectName</Name>
            <JaxbValueItem type="String">
                <StringValue>IDataSelectSettings.OBJECT_NAME</StringValue>
            </JaxbValueItem>
        </JaxbSettingValue>
        <JaxbSettingValue>
            <Name>PointNum</Name>
            <JaxbValueItem type="Integer">
                <IntValue>5</IntValue>
            </JaxbValueItem>
        </JaxbSettingValue>
    </JaxbAnalysisSettings>
    <JaxbAnalysisSettings>
        <Type>com.apldbio.sds.platform.core.analysis.ICrtSettings</Type>
        <JaxbSettingValue>
            <Name>CrtStartValue</Name>
            <JaxbValueItem type="Integer">
                <IntValue>1</IntValue>
            </JaxbValueItem>
        </JaxbSettingValue>
    </JaxbAnalysisSettings>
    <AnalysisSetting>
        <Type>NHC</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>LESS_THAN 35.0</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>BPR</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>GREATER_THAN 0.6</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>DRNMIN</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>LESS_THAN 35.0</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>LESS_THAN 0.2</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>FOS</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>true</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>HSD</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>GREATER_THAN 0.5</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>CC</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>LESS_THAN 0.8</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>NA</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>LESS_THAN 0.1</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>HRN</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>GREATER_THAN 4.0</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>NS</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>GREATER_THAN 1.0</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>EW</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>true</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria4QS</Name>
            <Type>String</Type>
            <StringValue>LESS_THAN 100000.0</StringValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>ORG</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>true</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>EAF</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>true</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>BAF</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>true</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>TAF</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>true</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>CAF</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>true</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>ROXLOW</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>LESS_THAN 20.0</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
    <AnalysisSetting>
        <Type>ROXDROP</Type>
        <SettingValue>
            <Name>enabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>true</BooleanValue>
        </SettingValue>
        <SettingValue>
            <Name>criteria</Name>
            <Type>String</Type>
            <StringValue>GREATER_THAN 0.04</StringValue>
        </SettingValue>
        <SettingValue>
            <Name>omitEnabled</Name>
            <Type>Boolean</Type>
            <BooleanValue>false</BooleanValue>
        </SettingValue>
    </AnalysisSetting>
</JaxbAnalysisProtocol>
"""



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

"""SCPI Command class and parsing"""

from __future__ import annotations

import re
import shlex
import textwrap
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Literal, Sequence, Type, TypeVar, cast

import numpy as np
import pyparsing as pp
from pyparsing import ParserElement
from pyparsing import pyparsing_common as ppc

pp.ParserElement.setDefaultWhitespaceChars("")

_ws_or_end = (
    (pp.Regex(r"[ \t\r]+") | pp.StringEnd() | pp.FollowedBy("\n"))
    .suppress()
    .setName("<ws/end>")
)
_nl = (
    (pp.Literal("\n") + pp.Optional(pp.Regex(r"[ \t\r]+")))
    .suppress()
    .setName("<newline>")
)
_fwe = pp.FollowedBy(_ws_or_end).suppress().setName("<fwe?>")
_fweqc = pp.FollowedBy(_ws_or_end | "<" | ",").suppress().setName("<fweqc?>")


def _make_multi_keyword(kwd_str: str, kwd_value: Any) -> ParserElement:
    x = pp.oneOf(kwd_str)
    x.setParseAction(pp.replaceWith(kwd_value))
    return x


_pbool = _make_multi_keyword("true True", True) | _make_multi_keyword(
    "False false", False
)

_qs = pp.quotedString
_qs.setParseAction(lambda toks: toks[0][1:-1])

_quote_content = pp.Word(pp.alphanums + "._")
_quote_open = pp.Combine("<" + _quote_content + ">")
_quote_close = pp.Combine("</" + pp.matchPreviousExpr(_quote_content) + ">")
_quote_close_any = pp.Combine("</" + _quote_content + ">")

_command_forward = pp.Forward()

_commands_block = (
    _quote_open.suppress()
    + _nl
    + pp.delimitedList(_command_forward, _nl)
    + _nl
    + _quote_close.suppress()
)

_opt_value_one = (
    (_pbool + _fweqc)
    | (ppc.number + _fweqc)
    | (_quote_open.suppress() + pp.Regex(r"[^<]+") + _quote_close.suppress() + _fweqc)
    | (pp.Regex(r"[^ \t\n<\",#']+") + _fweqc)
    | (_qs + _fweqc)
).setParseAction(lambda toks: toks[0])

_ovcl = pp.delimited_list(_opt_value_one, ",")

_opt_value = (_commands_block | _ovcl | _opt_value_one).setParseAction(
    lambda toks: toks[0]
    if not isinstance(toks[0], SCPICommand) and len(toks) == 1
    else (list(toks[:]),)
)

_opt_kv_pair = (
    pp.Literal("-").suppress()
    + ppc.identifier("key")
    + pp.Literal("=").suppress()
    + _opt_value.setResultsName("value")
).setResultsName("opt", listAllMatches=True)

_arg = _opt_value.setResultsName("arg", listAllMatches=True)

_arglist = pp.delimitedList(
    _opt_kv_pair
    | (_quote_open.suppress() + _opt_kv_pair + _quote_close.suppress())
    | _arg,
    _ws_or_end,
)

_arglist.setParseAction(
    lambda toks: ArgList(
        {k: v for k, v in toks.get("opt", [])}, list(toks.get("arg", []))
    )
)


@dataclass
class ArgList:
    "A representation of an SCPI list of options (-key=value) and arguments."
    opts: dict[str, bool | int | float | str]
    args: list[bool | int | float | str]

    @classmethod
    def from_string(cls, argument_string: str) -> ArgList:
        """Parse an SCPI argument string."""
        return cast(ArgList, _arglist.parseString(argument_string)[0])


_NULLIST = ArgList({}, [])

_commentstring: pp.ParserElement = pp.Combine(
    pp.Regex(r"\s*#\s?").suppress()
    + pp.Regex(r"[^\n]+")
    + (pp.StringEnd() | pp.FollowedBy("\n")).suppress()
)

_command: pp.ParserElement = cast(
    pp.ParserElement,
    (
        pp.Combine(pp.Regex("[A-Za-z:_]+"))("command")
        + _ws_or_end
        + pp.Optional(_arglist("arglist"))
        + pp.Optional(_commentstring("comment"))
        + _ws_or_end
    ).setParseAction(
        lambda toks: SCPICommand(
            toks["command"],
            comment=toks.get("comment", None),
            *toks.get("arglist", _NULLIST).args,  # FIXME
            **toks.get("arglist", _NULLIST).opts,
        )
    ),
)

_command_forward << _command


_accesslevel_order: dict[str, int] = {
    "Guest": 0,
    "Observer": 1,
    "Controller": 2,
    "Administrator": 3,
    "Full": 4,
}


class AccessLevel(Enum):
    """QS machine access level, with comparisons."""

    Guest = "Guest"
    Observer = "Observer"
    Controller = "Controller"
    Administrator = "Administrator"
    Full = "Full"
    # value: str

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, AccessLevel):
            other = AccessLevel(other)
        return _accesslevel_order[self.value] > _accesslevel_order[other.value]

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, AccessLevel):
            other = AccessLevel(other)
        return _accesslevel_order[self.value] >= _accesslevel_order[other.value]

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, AccessLevel):
            other = AccessLevel(other)
        return _accesslevel_order[self.value] < _accesslevel_order[other.value]

    def __le__(self, other: object) -> bool:
        if not isinstance(other, AccessLevel):
            other = AccessLevel(other)
        return _accesslevel_order[self.value] <= _accesslevel_order[other.value]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AccessLevel):
            other = AccessLevel(other)
        return _accesslevel_order[self.value] == _accesslevel_order[other.value]

    def __hash__(self) -> int:
        return self.value.__hash__()

    def __str__(self) -> str:
        return self.value


T = TypeVar("T")


class SCPICommandLike(ABC):
    """Abstract class for an object that can be converted from/to an SCPICommand."""

    @abstractmethod
    def to_scpicommand(self, **kwargs: Any) -> SCPICommand:  # pragma: no cover
        """Convert the object to an :any:`SCPICommand`"""
        ...

    @classmethod
    @abstractmethod
    def from_scpicommand(cls: Type[T], com: SCPICommand) -> T:  # pragma: no cover
        """Try to create the object from an :any:`SCPICommand`."""
        ...


@dataclass(init=False)
class SCPICommand(SCPICommandLike):
    """
    A representation of an SCPI Command.
    """

    command: str
    args: Sequence[
        str
        | int
        | float
        | np.number[Any]
        | Sequence[str | int | float | np.number[Any]]
        | Sequence["SCPICommand"]
    ]
    opts: dict[
        str,
        str
        | int
        | float
        | np.number[Any]
        | Sequence[str | int | float | np.number[Any]],
    ]
    comment: str | None

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, SCPICommand):
            return False

        return (
            (self.command == other.command)
            and (self.args == other.args)
            and (self.opts == other.opts)
        )

    def __init__(
        self,
        command: str,
        *args: str
        | int
        | float
        | np.number[Any]
        | Sequence[str | int | float | np.number[Any]]
        | Sequence["SCPICommand"],
        comment: str | None = None,
        **kwargs: str
        | int
        | float
        | np.number[Any]
        | Sequence[str | int | float | np.number[Any]],
    ) -> None:
        if " " in command:
            if args or comment or kwargs:
                raise ValueError
            n = SCPICommand.from_string(command)
            self.command = n.command
            self.args = n.args
            self.opts = n.opts
            self.comment = n.comment
            return

        self.command = command.upper()
        self.args = args
        self.opts = {k.lower(): v for k, v in kwargs.items()}
        self.comment = comment

    def _optformat(
        self,
        opt_val: str
        | int
        | float
        | np.number[Any]
        | Sequence[str | int | float | np.number[Any]]
        | Sequence[SCPICommand]
        | SCPICommand,
    ) -> str:
        if isinstance(opt_val, SCPICommand):
            opt_val = [opt_val]
        if isinstance(opt_val, str):
            if "\n" in opt_val:
                return f"<quote>{opt_val}</quote>"
            return shlex.quote(opt_val)
        if isinstance(opt_val, (int, float, np.number)):
            return str(opt_val)
        if isinstance(opt_val, (Sequence, np.ndarray)):
            if isinstance(opt_val[0], SCPICommand):
                q = "multiline." + self.command.lower()
                return (
                    f"<{q}>\n"
                    + textwrap.indent("".join(str(x) for x in opt_val), "\t")
                    + f"</{q}>"
                )
            return ",".join(self._optformat(x) for x in opt_val)

        raise TypeError(f"{opt_val}, of type {type(opt_val)}, not understood.")

    def to_string(self) -> str:
        """Create a usable command string, including terminal newline."""
        if self.comment:
            comment = f" # {self.comment}"
        else:
            comment = ""
        return (
            " ".join(
                [self.command]
                + [f"-{k}={self._optformat(v)}" for k, v in self.opts.items()]
                + [self._optformat(v) for v in self.args]
            )
            + comment
            + "\n"
        )

    @classmethod
    def from_string(cls, command_string: str) -> SCPICommand:
        """Parse (as SCPICommands) an SCPI command string."""
        return cast(SCPICommand, _command.parseString(command_string)[0])

    def specialize(self) -> SCPICommandLike:
        """If possible, convert SCPICommand to QSLib classes for the command."""
        if self.command.upper() in _scpi_command_classes:
            return _scpi_command_classes[self.command.upper()].from_scpicommand(self)
        return self

    def __str__(self) -> str:
        return self.to_string()

    def to_scpicommand(self, **kwargs: Any) -> SCPICommand:
        return self

    @classmethod
    def from_scpicommand(cls, com: SCPICommand) -> SCPICommand:
        return com


_scpi_command_classes: dict[str, SCPICommandLike] = {}



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
# SPDX-License-Identifier: AGPL-3.0-only

import pytest

from qslib import *


def test_create():

    exp = Experiment(protocol=Protocol([Stage([Step(30, 25)])]))


def test_fail_plots():
    exp = Experiment(protocol=Protocol([Stage([Step(30, 25)])]))

    with pytest.raises(ValueError, match="no temperature data"):
        exp.plot_temperatures()

    with pytest.raises(ValueError, match="no data available"):
        exp.plot_over_time()

    with pytest.raises(ValueError, match="no data available"):
        exp.plot_anneal_melt()



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("qslib")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = "dev"



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import annotations

import asyncio
import base64
import hmac
import io
import logging
import re
import shlex
import xml.etree.ElementTree as ET
import zipfile
from dataclasses import dataclass
from typing import Any, Dict, List, Literal, Optional, Union, cast, overload

import pandas as pd

from . import data
from .qs_is_protocol import CommandError, Error, NoMatch, QS_IS_Protocol
from .scpi_commands import AccessLevel, ArgList, SCPICommand

log = logging.getLogger(__name__)


def _gen_auth_response(password: str, challenge_string: str) -> str:
    return hmac.digest(password.encode(), challenge_string.encode(), "md5").hex()


def _parse_argstring(argstring: str) -> Dict[str, str]:
    unparsed = argstring.split()

    args: dict[str, str] = dict()
    # FIXME: do quotes allow spaces?
    for u in unparsed:
        m = re.match("-([^=]+)=(.*)$", u)
        if m is None:
            raise ValueError(f"Can't parse {u} in argstring.", u)
        args[m[1]] = m[2]

    return args


class AlreadyCollectedError(Exception):
    ...


class RunNotFinishedError(Exception):
    ...


@dataclass(frozen=True, order=True, eq=True)
class FilterDataFilename:
    filterset: data.FilterSet
    stage: int
    cycle: int
    step: int
    point: int

    @classmethod
    def fromstring(cls, x: str) -> FilterDataFilename:
        s = re.search(r"S(\d+)_C(\d+)_T(\d+)_P(\d+)_M(\d)_X(\d)_filterdata.xml$", x)
        if s is None:
            raise ValueError
        return cls(
            data.FilterSet.fromstring(f"x{s[6]}-m{s[5]}"),
            int(s[1]),
            int(s[2]),
            int(s[3]),
            int(s[4]),
        )

    def tostring(self) -> str:
        return f"S{self.stage:02}_C{self.cycle:03}_T{self.step:02}_P{self.point:04}_M{self.filterset.em}_X{self.filterset.ex}_filterdata.xml"

    def is_same_point(self, other: FilterDataFilename) -> bool:
        return (
            (self.stage == other.stage)
            and (self.cycle == other.cycle)
            and (self.step == other.step)
            and (self.point == other.point)
        )


class QSConnectionAsync:
    """Class for connection to a QuantStudio instrument server, using asyncio"""

    _protocol: QS_IS_Protocol

    async def __aenter__(self) -> QSConnectionAsync:
        await self.connect()
        return self

    async def __aexit__(self, exc_type: type, exc: Error, tb: Any) -> None:
        if self._transport.is_closing():
            return
        await self.disconnect()

    async def disconnect(self) -> None:
        if self._transport.is_closing():
            return
        await self._protocol.disconnect()
        self._transport.close()

    @property
    def connected(self) -> bool:
        if hasattr(self, "_protocol"):
            if self._protocol.lostconnection.done():
                return False
            else:
                return True
        else:
            return False

    @overload
    async def list_files(
        self,
        path: str,
        *,
        leaf: str = "FILE",
        verbose: Literal[True],
        recursive: bool = False,
    ) -> list[dict[str, Any]]:
        ...

    @overload
    async def list_files(
        self,
        path: str,
        *,
        leaf: str = "FILE",
        verbose: Literal[False],
        recursive: bool = False,
    ) -> list[str]:
        ...

    @overload
    async def list_files(
        self,
        path: str,
        *,
        leaf: str = "FILE",
        verbose: bool = False,
        recursive: bool = False,
    ) -> list[str] | list[dict[str, Any]]:
        ...

    async def list_files(
        self,
        path: str,
        *,
        leaf: str = "FILE",
        verbose: bool = False,
        recursive: bool = False,
    ) -> list[str] | list[dict[str, Any]]:
        if not verbose:
            if recursive:
                raise NotImplementedError
            return (await self.run_command(f"{leaf}:LIST? {path}")).split("\n")[1:-1]
        else:
            v = (await self.run_command(f"{leaf}:LIST? -verbose {path}")).split("\n")[
                1:-1
            ]
            ret: list[dict[str, str | float | int]] = []
            for x in v:
                rm = re.match(
                    r'"([^"]+)" -type=(\S+) -size=(\S+) -mtime=(\S+) -atime=(\S+) -ctime=(\S+)$',
                    x,
                )
                if rm is None:
                    ag = ArgList.from_string(x)
                    d: dict[str, str | float | int] = {}
                    d["path"] = ag.args[0]
                    d |= ag.opts
                else:
                    d = {}
                    d["path"] = rm.group(1)
                    d["type"] = rm.group(2)
                    d["size"] = int(rm.group(3))
                    d["mtime"] = float(rm.group(4))
                    d["atime"] = float(rm.group(5))
                    d["ctime"] = float(rm.group(6))
                if d["type"] == "folder" and recursive:
                    ret += await self.list_files(
                        cast(str, d["path"]), leaf=leaf, verbose=True, recursive=True
                    )
                else:
                    ret.append(d)
            return ret

    async def compile_eds(self, run_name: str) -> None:
        """Take a finished run directory in experiments:, compile it into an EDS, and move it to
        public_run_complete:"""

        expfiles = await self.list_files("", leaf="experiment", verbose=True)

        results = [r for r in expfiles if r["path"] == run_name]

        if len(results) != 1:
            raise ValueError
        res = results[0]

        if "run" not in res:
            raise FileNotFoundError(res)

        if res["state"] not in ["Completed", "Terminated"]:
            raise RunNotFinishedError(res)

        if ("collected" in res) and (res["collected"]):
            raise AlreadyCollectedError(res)

        await self.run_command(
            f'exp:run -asynchronous <block> zip "{run_name}.eds" "{run_name}" </block>'
        )

        await self.run_command(
            f'file:move "experiments:{run_name}.eds" "public_run_complete:{run_name}.eds"'
        )

        await self.run_command(f'exp:attr= "{run_name}" collected True')

    def __init__(
        self,
        host: str = "localhost",
        port: int = 7000,
        authenticate_on_connect: bool = True,
        initial_access_level: AccessLevel = AccessLevel.Observer,
        password: Optional[str] = None,
    ):
        """Create a connection to a QuantStudio Instrument Server."""
        self.host = host
        self.port = port
        self.password = password
        self._initial_access_level = initial_access_level
        self._authenticate_on_connect = authenticate_on_connect

    def _parse_access_line(self, aline: str) -> None:
        # pylint: disable=attribute-defined-outside-init
        if not aline.startswith("READy"):
            raise ConnectionError(f"Server opening seems invalid: {aline}")
        args = _parse_argstring(aline[5:])
        self.session = int(args["session"])
        self.product = args["product"]
        self.server_version = args["version"]
        self.server_build = args["build"]
        self.server_capabilities = args["capabilities"]
        self.server_hello_args = args

    async def connect(
        self,
        authenticate: Optional[bool] = None,
        initial_access_level: AccessLevel | None = None,
        password: Optional[str] = None,
    ) -> str:

        if authenticate is not None:
            self._authenticate_on_connect = authenticate
        if password is not None:
            self.password = password
        if initial_access_level is not None:
            self._initial_access_level = initial_access_level

        self.loop = asyncio.get_running_loop()
        self._transport, proto = await self.loop.create_connection(
            QS_IS_Protocol, self.host, self.port
        )

        self._protocol = cast(QS_IS_Protocol, proto)

        await self._protocol.readymsg
        resp = self._protocol.readymsg.result()
        self._parse_access_line(resp)

        if self._authenticate_on_connect:
            if self.password is not None:
                await self.authenticate(self.password)

        if self._initial_access_level is not None:
            await self.set_access_level(self._initial_access_level)

        return resp

    async def run_command_to_bytes(
        self, command: str | bytes | SCPICommand, just_ack: bool = True
    ) -> bytes:
        try:
            return (
                await self._protocol.run_command(command, just_ack=just_ack)
            ).rstrip()
        except CommandError as e:
            e.__traceback__ = None
            raise e

    async def run_command(
        self, command: str | bytes | SCPICommand, just_ack: bool = False
    ) -> str:
        try:
            return (await self.run_command_to_bytes(command, just_ack)).decode()
        except CommandError as e:
            e.__traceback__ = None
            raise e

    async def authenticate(self, password: str) -> None:
        challenge_key = await self.run_command(SCPICommand("CHAL?"))
        auth_rep = _gen_auth_response(password, challenge_key)
        await self.run_command(SCPICommand("AUTH", auth_rep))

    async def set_access_level(self, level: AccessLevel) -> None:
        await self.run_command(SCPICommand("ACC", level.value))

    async def get_expfile_list(
        self, glob: str, allow_nomatch: bool = False
    ) -> List[str]:
        try:
            fl = await self.run_command(SCPICommand("EXP:LIST?", glob))
        except NoMatch as ce:
            if allow_nomatch:
                return []
            else:
                raise ce
        else:
            assert fl.startswith("<quote.reply>")
            assert fl.endswith("</quote.reply>")
            return fl.split("\n")[1:-1]

    async def get_run_title(self) -> str:
        return (await self.run_command("RUNTitle?")).strip('"')

    async def get_exp_file(
        self, path: str, encoding: Literal["plain", "base64"] = "base64"
    ) -> bytes:
        reply = await self.run_command_to_bytes(
            f"EXP:READ? -encoding={encoding} {shlex.quote(path)}"
        )
        assert reply.startswith(b"<quote>\n")
        assert reply.endswith(b"</quote>")
        r = reply[8:-8]
        if encoding == "base64":
            return base64.decodebytes(r)
        else:
            return r

    async def read_dir_as_zip(self, path: str, leaf: str = "FILE") -> zipfile.ZipFile:
        """Read a directory on the

        Parameters
        ----------
        path : str
            path on the machine
        leaf : str, optional
            leaf to use, by default "FILE"

        Returns
        -------
        zipfile.ZipFile
            the returned zip file
        """
        x = await self.run_command_to_bytes(f'{leaf}:ZIPREAD? "{path}"')

        return zipfile.ZipFile(io.BytesIO(base64.decodebytes(x[7:-8])))

    async def read_file(
        self,
        path: str,
        context: str | None = None,
        leaf: str = "FILE",
        encoding: Literal["plain", "base64"] = "base64",
    ) -> bytes:
        if not context:
            contexts = ""
        elif context[-1] == ":":
            contexts = context
        else:
            contexts = context + ":"
        reply = await self.run_command_to_bytes(
            SCPICommand(f"{leaf}:READ?", contexts + path, encoding=encoding)
        )
        assert reply.startswith(b"<quote>\n")
        assert reply.endswith(b"</quote>")
        r = reply[8:-8]
        if encoding == "base64":
            return base64.decodebytes(r)
        else:
            return r

    async def get_sds_file(
        self,
        path: str,
        runtitle: Optional[str] = None,
        encoding: Literal["base64", "plain"] = "base64",
    ) -> bytes:
        if runtitle is None:
            runtitle = await self.get_run_title()
        return await self.get_exp_file(f"{runtitle}/apldbio/sds/{path}", encoding)

    async def get_run_start_time(self) -> float:
        return float(await self.run_command("RET ${RunStartTime:--}"))

    @overload
    async def get_filterdata_one(
        self,
        ref: FilterDataFilename,
        *,
        run: Optional[str] = None,
        return_files: Literal[True],
    ) -> tuple[data.FilterDataReading, list[tuple[str, bytes]]]:
        ...

    @overload
    async def get_filterdata_one(
        self,
        ref: FilterDataFilename,
        *,
        run: Optional[str] = None,
        return_files: Literal[False] = False,
    ) -> data.FilterDataReading:
        ...

    async def get_filterdata_one(
        self,
        ref: FilterDataFilename,
        *,
        run: Optional[str] = None,
        return_files: bool = False,
    ) -> data.FilterDataReading | tuple[
        data.FilterDataReading, list[tuple[str, bytes]]
    ]:
        if run is None:
            run = await self.get_run_title()

        fl = await self.get_exp_file(f"{run}/apldbio/sds/filter/" + ref.tostring())

        if (x := ET.parse(io.BytesIO(fl)).find("PlatePointData/PlateData")) is not None:
            f = data.FilterDataReading(x)
        else:
            raise ValueError("PlateData not found")

        ql = (
            await self.get_expfile_list(
                f"{run}/apldbio/sds/quant/{f.filename_reading_string}_E*.quant"
            )
        )[-1]
        qf = await self.get_exp_file(ql)

        f.set_timestamp_by_quantdata(qf.decode())

        if return_files:
            files = [("filter/" + ref.tostring(), fl)]
            qn = re.search("quant/.*$", ql)
            assert qn is not None
            files.append((qn[0], qf))
            return f, files
        else:
            return f

    @overload
    async def get_all_filterdata(
        self, run: Optional[str], as_list: Literal[True]
    ) -> List[data.FilterDataReading]:
        ...

    @overload
    async def get_all_filterdata(
        self, run: Optional[str], as_list: Literal[False]
    ) -> pd.DataFrame:
        ...

    async def get_all_filterdata(
        self, run: str | None = None, as_list: bool = False
    ) -> Union[pd.DataFrame, List[data.FilterDataReading]]:
        if run is None:
            run = await self.get_run_title()

        pl = [
            await self.get_filterdata_one(FilterDataFilename.fromstring(x))
            for x in await self.get_expfile_list(
                f"{run}/apldbio/sds/filter/*_filterdata.xml"
            )
        ]

        if as_list:
            return pl

        return data.df_from_readings(pl)



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import annotations

import asyncio
import functools
import logging
import os
import re
import shutil
import time
import zipfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, TextIO, Tuple, Type, Union, cast

import numpy as np
import numpy.typing as npt
from influxdb_client import InfluxDBClient, Point  # , Point, WritePrecision
from influxdb_client.client.write_api import ASYNCHRONOUS
from nio.client import AsyncClient
from nio.client.async_client import AsyncClientConfig
from nio.responses import JoinedRoomsError

from qslib.base import RunStatus
from qslib.plate_setup import PlateSetup
from qslib.qs_is_protocol import CommandError
from qslib.qsconnection_async import FilterDataFilename, QSConnectionAsync
from qslib.scpi_commands import AccessLevel, ArgList

log = logging.getLogger("monitor")

LEDSTATUS = re.compile(
    rb"Temperature:([+\-\d.]+) Current:([+\-\d.]+) Voltage:([+\-\d.]+) JuncTemp:([+\-\d.]+)"
)


@dataclass
class LEDStatus:
    temperature: float
    current: float
    voltage: float
    junctemp: float


@dataclass(frozen=True)
class MatrixConfig:
    password: str
    user: str
    room: str
    host: str
    encryption: bool = False


@dataclass(frozen=True)
class InfluxConfig:
    token: str
    org: str
    bucket: str
    url: str


@dataclass(frozen=True)
class MachineConfig:
    password: Union[str, None] = None
    name: str = "localhost"
    host: str = "localhost"
    port: str = "7000"
    retries: int = 3
    compile: bool = False


@dataclass(frozen=True)
class SyncConfig:
    completed_directory: Union[str, None] = None
    in_progress_directory: Union[str, None] = None


@dataclass(frozen=True)
class Config:
    matrix: Union[MatrixConfig, None] = None
    influxdb: Union[InfluxConfig, None] = None
    machine: MachineConfig = MachineConfig()
    sync: SyncConfig = SyncConfig()


@dataclass
class RunState:
    name: Optional[str] = None
    stage: Optional[int | str] = None
    cycle: Optional[int] = None
    step: Optional[int] = None
    plate_setup: Optional[PlateSetup] = None

    async def refresh(self, c: QSConnectionAsync) -> None:
        runmsg = ArgList.from_string(await c.run_command("RunProgress?"))
        name = cast(str, runmsg.opts["RunTitle"])
        if name == "-":
            self.name = None
        else:
            self.name = re.sub(r"(<([\w.]+)>)?([^<]+)(</[\w.]+>)?", r"\3", name)
        stage = runmsg.opts["Stage"]
        if stage == "-":
            self.stage = None
        else:
            self.stage = cast(int, stage)
        cycle = runmsg.opts["Cycle"]
        if cycle == "-":
            self.cycle = None
        else:
            self.cycle = cast(int, cycle)
        step = runmsg.opts["Step"] if self.stage else None
        if step == "-":
            self.step = None
        else:
            self.step = cast(Optional[int], step)
        if self.name:
            try:
                self.plate_setup = await PlateSetup.from_machine(c)
            except CommandError:
                self.plate_setup = None

    @classmethod
    async def from_machine(cls: Type[RunState], c: QSConnectionAsync) -> RunState:
        n = cls.__new__(cls)
        await n.refresh(c)
        return n

    def statemsg(self, timestamp: str) -> str:
        s = f'run_state name="{self.name}"'
        if self.stage:
            s += f",stage={self.stage}i,cycle={self.cycle}i,step={self.step}i"
        else:
            s += f",stage=0i,cycle=0i,step=0i"  # FIXME: not great
        s += f" {timestamp}"
        return s


@dataclass
class MachineState:
    zone_targets: List[float]
    zone_controls: List[bool]
    cover_target: float
    cover_control: bool
    drawer: str

    async def refresh(self, c: QSConnectionAsync) -> None:
        targmsg = ArgList.from_string(await c.run_command("TBC:SETT?"))
        self.cover_target = cast(float, targmsg.opts["Cover"])
        self.zone_targets = cast(
            List[float], [targmsg.opts[f"Zone{i}"] for i in range(1, 7)]
        )

        contmsg = ArgList.from_string(await c.run_command("TBC:CONT?"))
        self.cover_control = cast(bool, contmsg.opts["Cover"])
        self.zone_controls = cast(
            List[bool], [contmsg.opts[f"Zone{i}"] for i in range(1, 7)]
        )

        self.drawer = await c.run_command("DRAW?")

    @classmethod
    async def from_machine(cls, c: QSConnectionAsync) -> MachineState:
        n = cast(MachineState, cls.__new__(cls))
        await n.refresh(c)
        return n

    # def targetmsg(timestamp):
    #    return ""


@dataclass
class State:
    run: RunState
    machine: MachineState

    @classmethod
    async def from_machine(cls, c: QSConnectionAsync) -> State:
        run = await RunState.from_machine(c)
        machine = await MachineState.from_machine(c)
        return cls(run, machine)


# def parse_fd_fn(x: str) -> Tuple[str, int, int, int, int]:
#    s = re.search(r"S(\d{2})_C(\d{3})_T(\d{2})_P(\d{4})_M(\d)_X(\d)_filterdata.xml$", x)
#    assert s is not None
#    return (f"x{s[6]}-m{s[5]}", int(s[1]), int(s[2]), int(s[3]), int(s[4]))


def index_to_filename_ref(i: Tuple[str, int, int, int, int]) -> str:
    x, s, c, t, p = i
    return f"S{s:02}_C{c:03}_T{t:02}_P{p:04}_M{x[4]}_X{x[1]}"


async def get_runinfo(c: QSConnectionAsync) -> State:
    state = await State.from_machine(c)
    return state


class Collector:
    def __init__(self, config: Config):
        self.config = config

        if self.config.influxdb:
            self.idbclient = InfluxDBClient(
                url=self.config.influxdb.url,
                token=self.config.influxdb.token,
                org=self.config.influxdb.org,
            )
            self.idbw = self.idbclient.write_api(write_options=ASYNCHRONOUS)
        else:
            self.idbw = None

        if self.config.matrix:
            self.matrix_config = AsyncClientConfig(
                encryption_enabled=self.config.matrix.encryption
            )

            self.matrix_client = AsyncClient(
                self.config.matrix.host,
                self.config.matrix.user,
                store_path="./matrix_store/",
                config=self.matrix_config,
            )

        log.info(config.sync)

        self.run_log_file: TextIO | None = None

    def inject(
        self, t: str | Iterable[str | Point] | Point, flush: bool = False
    ) -> None:
        if self.idbw:
            self.idbw.write(bucket=self.config.influxdb.bucket, record=t)  # type:ignore
            if flush:
                self.idbw.flush()
        else:
            pass

    async def matrix_announce(self, msg: str) -> None:
        assert self.config.matrix
        await self.matrix_client.room_send(
            room_id=self.config.matrix.room,
            message_type="m.room.message",
            content={"msgtype": "m.text", "body": msg},
            ignore_unverified_devices=True,
        )

        await self.matrix_client.sync()

    async def setup_new_rundir(
        self,
        connection: QSConnectionAsync,
        name: str,
        *,
        firstmsg: str | None = None,
        overwrite: bool = False,
    ) -> None:
        # name = name.replace(" ", "_")

        assert self.ipdir is not None

        if not self.ipdir.is_dir():
            log.error(f"Can't open in-progress directory {self.ipdir}.")
            return

        dirpath = self.ipdir / name

        if dirpath.exists() and (not overwrite):
            log.error(f"In-progress directory for {name} already exists.")
            return
        elif dirpath.exists():
            assert dirpath != self.ipdir
            shutil.rmtree(dirpath)

        dirpath.mkdir()
        zf = await connection.read_dir_as_zip(name, "experiment")
        zf.extractall(dirpath)

        (dirpath / "apldbio" / "sds" / "quant").mkdir(exist_ok=True)
        (dirpath / "apldbio" / "sds" / "filter").mkdir(exist_ok=True)
        (dirpath / "apldbio" / "sds" / "calibrations").mkdir(exist_ok=True)

        self.run_log_file = (dirpath / "apldbio" / "sds" / "messages.log").open("a")

        if firstmsg is not None:
            self.run_log_file.write(firstmsg)
            self.run_log_file.flush()

    @property
    def ipdir(self) -> Path | None:
        x = self.config.sync.in_progress_directory
        if x is None:
            return None
        else:
            return Path(x)

    def run_ip_path(self, name: str) -> Path:
        # name = name.replace(" ", "_")
        if (ipdir := self.ipdir) is None:
            raise ValueError
        return ipdir / name / "apldbio" / "sds"

    async def compile_eds(self, connection: QSConnectionAsync, name: str) -> None:
        # name = name.replace(" ", "_")

        # Wait 5 minutes in case machine compiles it (AB sofware run)
        await asyncio.sleep(300.0)

        try:
            await connection.set_access_level(AccessLevel.Controller)
            await connection.compile_eds(name)
        except FileNotFoundError as e:
            raise e
        finally:
            await connection.set_access_level(AccessLevel.Observer)

    async def sync_completed(self, connection: QSConnectionAsync, name: str) -> None:
        # name = name.replace(" ", "_")

        try:
            await self.compile_eds(connection, name)
        except FileNotFoundError:
            pass

        dir = Path(cast(str, self.config.sync.completed_directory))

        if not dir.is_dir():
            log.error(f"Can't sync completed EDS to invalid path {dir}.")
            return

        path = dir / (name + ".eds")

        if path.exists():
            log.error(f"Completed EDS already exists for {name}.")
            return

        try:
            with path.open("wb") as f:
                edsfile = await connection.read_file(f"public_run_complete:{name}.eds")
                f.write(edsfile)
        except Exception as e:
            log.error(f"Error synchronizing completed EDS {name}: {e}")
            return

        if self.ipdir:
            import shutil

            if (self.ipdir / name).exists():
                shutil.rmtree(self.ipdir / name)
            if (x := (self.ipdir / (name + ".eds"))).exists():
                x.unlink()

    async def docollect(
        self,
        args: Dict[str, Union[str, int, bool, float]],
        state: State,
        connection: QSConnectionAsync,
    ) -> None:
        if state.run.plate_setup:
            pa: npt.NDArray[
                np.object_
            ] | None = state.run.plate_setup.well_samples_as_array()
        else:
            pa = None

        run = cast(str, args["run"])

        if run.startswith('"'):
            run = run[1:-1]

        del args["run"]
        for k, v in args.items():
            if k != "run":
                args[k] = int(v)
        pl = [
            FilterDataFilename.fromstring(x)
            for x in await connection.get_expfile_list(
                "{run}/apldbio/sds/filter/S{stage:02}_C{cycle:03}"
                "_T{step:02}_P{point:04}_*_filterdata.xml".format(
                    run=run, **cast(Dict[str, int], args)
                ),
                allow_nomatch=True,
            )
        ]
        pl.sort()
        toget = [x for x in pl if x.is_same_point(pl[-1])]

        lp: List[str] = []
        files: list[tuple[str, bytes]] = []

        if (
            self.ipdir
            and (
                self.ipdir / run / "apldbio" / "sds" / "filter"  # .replace(" ", "_")
            ).exists()
        ):
            for fdf in toget:
                fdr, files_one = await connection.get_filterdata_one(
                    fdf, return_files=True
                )
                lp += fdr.to_lineprotocol(run_name=run, sample_array=pa)
                files += files_one
        else:
            for fdf in toget:
                lp += (await connection.get_filterdata_one(fdf)).to_lineprotocol(
                    run_name=run, sample_array=pa
                )

        self.inject(lp, flush=True)

        for path, data in files:
            fullpath = self.run_ip_path(run) / path
            with fullpath.open("wb") as f:
                f.write(data)

        if (
            self.ipdir
            and (
                self.ipdir / run / "apldbio" / "sds" / "filter"  # .replace(" ", "_")
            ).exists()
        ):
            saferun = run  # .replace(" ", "_")
            ipp = self.ipdir / saferun
            with zipfile.ZipFile(self.ipdir / (saferun + ".eds"), "w") as z:
                for root, _, zfiles in os.walk(ipp):
                    for zfile in zfiles:
                        fpath = os.path.join(root, zfile)
                        z.write(fpath, os.path.relpath(fpath, ipp))

    async def handle_run_msg(
        self: Collector,
        state: State,
        c: QSConnectionAsync,
        topic: bytes,
        message: bytes,
        timestamp: float | None,
    ) -> None:
        topic_str = topic.decode()
        message_str = message.decode()

        # Are we logging?
        if self.run_log_file is not None:
            self.run_log_file.write(f"{topic_str} {timestamp} {message_str}")
            self.run_log_file.flush()

        assert timestamp is not None
        timestamp = int(1e9 * timestamp)
        msg = ArgList.from_string(message_str)
        log.debug(msg)
        contents = msg.args
        action = cast(str, contents[0])
        if action == "Stage":
            assert isinstance(contents[1], (str, int))
            state.run.stage = contents[1]
            self.inject(
                Point("run_action")
                .tag("type", action.lower())
                .field(action.lower(), contents[1])
                .time(timestamp)
                .to_line_protocol()
            )
            self.inject(
                Point("run_status")
                .tag("type", action.lower())
                .field(action.lower(), contents[1])
                .time(timestamp)
                .to_line_protocol()
            )
        elif action == "Cycle":
            state.run.cycle = int(contents[1])
            self.inject(
                Point("run_action")
                .tag("type", action.lower())
                .field(action.lower(), contents[1])
                .time(timestamp)
                .to_line_protocol()
            )
            self.inject(
                Point("run_status")
                .tag("type", action.lower())
                .field(action.lower(), contents[1])
                .time(timestamp)
                .to_line_protocol()
            )
        elif action == "Step":
            state.run.step = int(contents[1])
            self.inject(
                Point("run_status")
                .tag("type", action.lower())
                .field(action.lower(), contents[1])
                .time(timestamp)
                .to_line_protocol()
            )
            self.inject(
                Point("run_action")
                .tag("type", action.lower())
                .field(action.lower(), contents[1])
                .time(timestamp)
                .to_line_protocol()
            )
        elif action == "Holding":
            self.inject(
                f'run_action,type=Holding holdtime={msg.opts["time"]} {timestamp}'  # noqa: E501
            )
        elif action == "Ramping":
            # TODO: check zones
            state.machine.zone_targets = [
                float(x) for x in cast(list[float], msg.opts["targets"][0])  # type: ignore
            ]
            self.inject(
                f'run_action,type={action} run_name="{state.run.name}" {timestamp}'  # noqa: E501
            )
        elif action == "Acquiring":
            self.inject(
                f'run_action,type={action} run_name="{state.run.name}" {timestamp}'  # noqa: E501
            )
        elif action in ["Error", "Ended", "Aborted", "Stopped", "Starting"]:
            self.inject(
                f'run_action,type={action} run_name="{state.run.name}" {timestamp}'  # noqa: E501
            )
            asyncio.tasks.create_task(
                self.matrix_announce(
                    f"{self.config.machine.name} status: {action} {' '.join(str(x) for x in contents[1:])}"  # noqa: E501
                )
            )
            if action == "Ended":
                self.run_log_file = None

                if self.config.machine.compile:
                    assert state.run.name
                    compdir = self.config.sync.completed_directory

                    if compdir != "":
                        # This will need to compile and sync
                        asyncio.tasks.create_task(
                            self.sync_completed(c, state.run.name)
                        )
                    else:
                        # No sync; just compile
                        asyncio.tasks.create_task(self.compile_eds(c, state.run.name))
            elif action == "Starting":
                if self.ipdir:
                    newname: str = cast(str, contents[1])
                    newname.strip('"')

                    asyncio.tasks.create_task(
                        self.setup_new_rundir(
                            c,
                            newname,
                            firstmsg=f"\n{topic_str} {timestamp} {message_str}",
                        )
                    )

        elif action == "Collected":
            self.inject(
                f'run_action,type={action} run_name="{state.run.name}" {timestamp}'  # noqa: E501
            )
            asyncio.tasks.create_task(self.docollect(msg.opts, state, c))
        else:
            self.inject(
                Point("run_action")
                .tag("type", f"Other")
                .tag("run_name", state.run.name)
                .field("message", " ".join(str(x) for x in contents))
                .time(timestamp)
            )

        await state.run.refresh(c)
        await state.machine.refresh(c)

        log.info(message_str)
        self.inject(state.run.statemsg(str(timestamp)))

        if state.run.plate_setup:
            self.inject(
                state.run.plate_setup.to_lineprotocol(timestamp, state.run.name)
            )

        if self.idbw:
            self.idbw.flush()

    async def handle_led(
        self, topic: bytes, message: bytes, timestamp: float | None
    ) -> None:
        # Are we logging?
        if self.run_log_file is not None:
            self.run_log_file.write(f"{topic.decode()} {timestamp} {message.decode()}")
            self.run_log_file.flush()
        assert timestamp is not None
        ls = LEDSTATUS.match(message)
        assert ls
        p = (
            Point("lamp")
            .field("temperature", float(ls[1].decode()))
            .field("current", float(ls[2].decode()))
            .field("voltage", float(ls[3].decode()))
            .field("junctemp", float(ls[4].decode()))
            .time(int(1e9 * timestamp))
        )
        self.inject(p, flush=True)

    async def handle_msg(
        self,
        state: State,
        c: QSConnectionAsync,
        topic: bytes,
        message: bytes,
        timestamp: float | None,
    ) -> None:
        # Are we logging?
        if self.run_log_file is not None:
            self.run_log_file.write(f"{topic.decode()} {timestamp} {message.decode()}")
            self.run_log_file.flush()
        assert timestamp is not None
        args = ArgList.from_string(message.decode()).opts
        log.debug(f"Handling message {topic.decode()} {message.decode()}")
        if topic == b"Temperature":
            recs = []
            for i, (s, b, t) in enumerate(
                zip(
                    # FIXME: parsing weirdness: these are single-element tuples
                    [float(x) for x in cast(list[float], args["sample"][0])],  # type: ignore
                    [float(x) for x in cast(list[float], args["block"][0])],  # type: ignore
                    state.machine.zone_targets,
                )
            ):
                recs.append(
                    f"temperature,loc=zones,zone={i} sample={s},block={b},target={t} {int(1e9 * timestamp)}"  # noqa: E501
                )
            recs.append(
                Point("temperature").tag("loc", "cover").field("cover", args["cover"])
            )
            recs.append(
                Point("temperature")
                .tag("loc", "heatsink")
                .field("heatsink", args["heatsink"])
            )
            self.inject(recs)
        elif topic == b"Time":
            p = Point("run_time")
            for key in ["elapsed", "remaining", "active"]:
                if key in args.keys():
                    p = p.field(key, args[key])
                p.time(int(1e9 * timestamp))
            self.inject(p)
        if self.idbw:
            self.idbw.flush()

    async def monitor(self, connected_fut: asyncio.Future[bool] | None = None) -> None:

        if self.config.matrix is not None:
            await self.matrix_client.login(self.config.matrix.password)
            joinedroomresp = await self.matrix_client.joined_rooms()
            if isinstance(joinedroomresp, JoinedRoomsError):
                log.error(joinedroomresp)
                joinedrooms = []
            else:
                joinedrooms = joinedroomresp.rooms
            if self.config.matrix.room not in joinedrooms:
                await self.matrix_client.join(self.config.matrix.room)

        async with QSConnectionAsync(
            host=self.config.machine.host,
            port=int(self.config.machine.port),
            password=self.config.machine.password,
        ) as c:
            log.info("monitor connected")
            # Are we currently *in* a run? If so, we'll need to get info.
            state = await get_runinfo(c)
            log.info(f"status info: {state}")

            self.inject(state.run.statemsg(str(time.time_ns())))

            if state.run.plate_setup:
                self.inject(
                    state.run.plate_setup.to_lineprotocol(
                        time.time_ns(), state.run.name
                    )
                )

            if self.idbw:
                self.idbw.flush()

            # Setup directory if run already started:
            if state.run.name and self.ipdir:
                await self.setup_new_rundir(c, state.run.name, overwrite=True)

            await c.run_command("SUBS -timestamp Temperature Time Run LEDStatus")
            log.debug("subscriptions made")

            for t in [b"Temperature", b"Time"]:
                c._protocol.topic_handlers[t] = functools.partial(
                    self.handle_msg, state, c
                )
            c._protocol.topic_handlers[b"Run"] = functools.partial(
                self.handle_run_msg, state, c
            )

            c._protocol.topic_handlers[b"LEDStatus"] = self.handle_led

            log.info(c._protocol.topic_handlers)

            if connected_fut is not None:
                connected_fut.set_result(True)

            ok = True
            while ok:
                await asyncio.wait((c._protocol.lostconnection,), timeout=60)

                # Have we lost the connection?
                if c._protocol.lostconnection.done():
                    log.error("Lost connection.")
                    ok = False

                # Are we actually fine?
                if time.time() - c._protocol.last_received <= 60.0:
                    continue

                # No, we have a sleep timeout.  Send a test command.
                try:
                    await asyncio.wait_for(c.run_command("ISTAT?"), 30.0)
                except TimeoutError:
                    log.error(
                        "No data received in 5 minutes and ISTAT? test timed out.  Trying to disconnect."
                    )
                    await c.disconnect()
                    raise TimeoutError

    async def reliable_monitor(
        self, connected_fut: asyncio.Future[bool] | None = None
    ) -> None:
        log.info("starting reconnectable monitoring")

        restart = True
        successive_failures = 0
        while restart:
            try:
                await self.monitor(connected_fut=connected_fut)
            except asyncio.exceptions.TimeoutError as e:
                successive_failures = 0
                log.warn(f"lost connection with timeout {e}")
            except OSError as e:
                log.error(f"connectio error {e}, retrying")
            except Exception as e:
                if self.config.machine.retries - successive_failures > 0:
                    log.error(
                        f"Error {repr(e)}, retrying {self.config.machine.retries-successive_failures} times"  # noqa: E501
                    )
                    successive_failures += 1
                else:
                    log.critical(f"giving up, error {e}")
                    restart = False
            log.debug("awaiting retry")
            await asyncio.sleep(30)



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import annotations

import asyncio
import io
import logging
import re
import time
from asyncio.futures import Future
from dataclasses import dataclass
from typing import Any, Coroutine, Optional, Protocol, Type

from .scpi_commands import AccessLevel, SCPICommand

NL_OR_Q = re.compile(rb"(?:\n|<(/?)([\w.]+)[ *]*>)")
Q_ONLY = re.compile(rb"<(/?)([\w.]+)[ *]*>")
TIMESTAMP = re.compile(rb"(\d{8,}\.\d{3})")

log = logging.getLogger(__name__)


def _validate_command_format(commandstring: bytes) -> None:
    # This is meant to validate that the command will not mess up comms
    # The command may be completely malformed otherwise

    tagseq: list[tuple[bytes, bytes, bytes]] = re.findall(
        rb"<(/?)([\w.]+)[ *]*>|(\n)", commandstring.rstrip()
    )  # tuple of close?,tag,newline?

    tagstack: list[bytes] = []
    for c, t, n in tagseq:
        if n and not tagstack:
            raise ValueError("newline outside of quotation")
        elif t and not c:
            tagstack.append(t)
        elif c:
            if not tagstack:
                raise ValueError(f"unbalanced tag <{c.decode()}{t.decode()}>")
            opentag = tagstack.pop()
            if opentag != t:
                raise ValueError(
                    f"unbalanced tags <{opentag.decode()}> <{c.decode()}{t.decode()}>"
                )
        elif n:
            continue
        else:
            raise ValueError("Unknown")
    if tagstack:
        raise ValueError("Unclosed tags")


class Error(Exception):
    pass


class CommandError(Error):
    @staticmethod
    def parse(command: str, ref_index: str, response: str) -> CommandError:
        m = re.match(r"\[(\w+)\] (.*)", response)
        if (not m) or (m[1] not in COM_ERRORS):
            return UnparsedCommandError(command, ref_index, response)
        try:
            return COM_ERRORS[m[1]].parse(command, ref_index, m[2])
        except ValueError:
            return UnparsedCommandError(command, ref_index, response)


@dataclass
class UnparsedCommandError(CommandError):
    command: Optional[str]
    ref_index: Optional[str]
    response: str


@dataclass
class InsufficientAccess(CommandError):
    command: str
    requiredAccess: AccessLevel
    currentAccess: AccessLevel
    message: str

    @classmethod
    def parse(cls, command: str, ref_index: str, message: str) -> InsufficientAccess:
        m = re.match(
            r'-requiredAccess="(\w+)" -currentAccess="(\w+)" --> (.*)', message
        )
        if not m:
            raise ValueError
        return cls(command, AccessLevel(m[1]), AccessLevel(m[2]), m[3])


@dataclass
class AuthError(CommandError):
    command: str
    message: str

    @classmethod
    def parse(cls, command: str, ref_index: str, message: str) -> AuthError:
        m = re.match(r"--> (.*)", message)
        if not m:
            raise ValueError
        return cls(command, m[1])


@dataclass
class AccessLevelExceeded(CommandError):
    command: str
    accessLimit: AccessLevel
    message: str

    @classmethod
    def parse(cls, command: str, ref_index: str, message: str) -> AccessLevelExceeded:
        m = re.match(r'-accessLimit="(\w+)" --> (.*)', message)
        if not m:
            raise ValueError
        return cls(command, AccessLevel(m[1]), m[2])


@dataclass
class InvocationError(CommandError):
    command: str
    message: str

    @classmethod
    def parse(cls, command: str, ref_index: str, message: str) -> InvocationError:
        return cls(command, message)


@dataclass
class NoMatch(CommandError):
    command: str
    message: str

    @classmethod
    def parse(cls, command: str, ref_index: str, message: str) -> NoMatch:
        return cls(command, message)


COM_ERRORS: dict[str, Type[CommandError]] = {
    "InsufficientAccess": InsufficientAccess,
    "AuthError": AuthError,
    "AccessLevelExceeded": AccessLevelExceeded,
    "InvocationError": InvocationError,
    "NoMatch": NoMatch,
}


class ReplyError(IOError):
    pass


class SubHandler(Protocol):
    def __call__(
        self, topic: bytes, message: bytes, timestamp: float | None
    ) -> Coroutine[None, None, None]:  # pragma: no cover
        ...


class QS_IS_Protocol(asyncio.Protocol):
    lostconnection: Future[Any]
    last_received: float
    waiting_commands: list[
        tuple[
            bytes,
            None
            | Future[tuple[bytes, bytes, None | Future[tuple[bytes, bytes, None]]]],
        ]
    ]

    def __init__(self) -> None:
        self.default_topic_handler = self._default_topic_handler
        self.readymsg: Future[str] = asyncio.get_running_loop().create_future()
        self.lostconnection = asyncio.get_running_loop().create_future()
        self.should_be_connected = False
        self.last_received = time.time()

    def connection_lost(self, exc: Optional[Exception]) -> None:
        if self.should_be_connected:
            log.warn("Lost connection")
        else:
            log.info("Connection closed")
        self.lostconnection.set_result(exc)
        self.should_be_connected = False

        # Cancel all futures; we'll never recover them.
        for _, future in self.waiting_commands:
            if future is not None:
                future.cancel()

        self.waiting_commands = []

    async def disconnect(self) -> None:
        self.should_be_connected = False
        await self.run_command("QUIT")

    def connection_made(self, transport: Any) -> None:
        log.info("Made connection")
        self.should_be_connected = True
        # setup connection.
        self.transport = transport
        self.waiting_commands = []
        self.buffer = io.BytesIO()
        self.quote_stack: list[bytes] = []
        self.topic_handlers: dict[bytes, SubHandler] = {}
        self.last_received = time.time()

    async def _default_topic_handler(
        self, topic: bytes, message: bytes, timestamp: Optional[float] = None
    ) -> None:
        log.info(f"{topic.decode()} at {timestamp}: {message.decode()}")
        self.last_received = time.time()

    async def handle_sub_message(self, message: bytes) -> None:
        i = message.index(b" ")
        topic = message[0:i]
        if m := TIMESTAMP.match(message, i + 1):
            timestamp: float | None = float(m[1])
            i = m.end()
        else:
            timestamp = None
        asyncio.create_task(
            self.topic_handlers.get(topic, self.default_topic_handler)(
                topic, message[i + 1 :], timestamp=timestamp
            )
        )

    async def parse_message(self, ds: bytes) -> None:
        if ds.startswith((b"ERRor", b"OK", b"NEXT")):
            ms = ds.index(b" ")
            r = None
            comfut_new = None
            if ds.startswith(b"NEXT"):
                loop = asyncio.get_running_loop()
                comfut_new = loop.create_future()
            for i, (commref, comfut) in enumerate(self.waiting_commands):
                if ds.startswith(commref, ms + 1):
                    if comfut is not None:
                        comfut.set_result(
                            (ds[:ms], ds[ms + len(commref) + 2 :], comfut_new)
                        )
                    else:
                        log.info(f"{commref!r} complete: {ds!r}")
                    r = i
                    break
            if r is None:
                log.error(f"received unexpected command response: {ds!r}")
            elif ds.startswith(b"NEXT"):
                self.waiting_commands.append((self.waiting_commands[r][0], comfut_new))
                del self.waiting_commands[r]
            else:
                del self.waiting_commands[r]
        elif ds.startswith(b"MESSage"):
            await self.handle_sub_message(ds[8:])
        elif ds.startswith(b"READy"):
            self.readymsg.set_result(ds.decode())
        else:
            log.error(f"Unknown message: {ds!r}")

    def data_received(self, data: bytes) -> None:
        """Process received data packet from instrument, keeping track of quotes. If
        a newline occurs when the quote stack is empty, create a task to process
        the message, but continue processing. (TODO: consider threads/processes here.)

        :param data: bytes:

        """
        lastwrite = 0
        for m in NL_OR_Q.finditer(data):
            if m[0] == b"\n":
                if len(self.quote_stack) == 0:
                    self.buffer.write(data[lastwrite : m.end()])
                    lastwrite = m.end()
                    asyncio.create_task(self.parse_message(self.buffer.getvalue()))
                    self.buffer = io.BytesIO()
                # else:  # This is not actually needed
                #     continue
            else:
                if not m[1]:
                    self.quote_stack.append(m[2])
                else:
                    try:
                        i = self.quote_stack.index(m[2])
                    except ValueError:
                        log.error(
                            f"Close quote {m[2]!r} did not have open in stack {self.quote_stack}. Disconnecting to avoid corruption."
                        )
                        self.quote_stack = []
                        self.connection_lost(ConnectionError())
                    else:
                        self.quote_stack = self.quote_stack[:i]
                self.buffer.write(data[lastwrite : m.end()])
                lastwrite = m.end()
        self.buffer.write(data[lastwrite:])
        self.last_received = time.time()

    async def run_command(
        self,
        comm: str | bytes | SCPICommand,
        ack_timeout: int = 300,
        just_ack: bool = True,
        uid: bool = True,
    ) -> bytes:
        if isinstance(comm, str):
            comm = comm.encode()
        elif isinstance(comm, SCPICommand):
            comm = comm.to_string().encode()
        comm = comm.rstrip()
        _validate_command_format(comm)
        log.debug(f"Running command {comm.decode()}")
        loop = asyncio.get_running_loop()

        comfut: Future[
            tuple[bytes, bytes, None | Future[tuple[bytes, bytes, None]]]
        ] = loop.create_future()
        if uid:
            import random

            commref = str(random.randint(1, 2**30)).encode()
            comm_with_ref = commref + b" " + comm
        else:
            comm_with_ref = comm
        self.transport.write((comm_with_ref + b"\n"))
        log.debug(f"Sent command {comm_with_ref!r}")
        if m := re.match(rb"^(\d+) ", comm_with_ref):
            commref = m[1]
        else:
            commref = comm
        self.waiting_commands.append((commref, comfut))

        try:
            await asyncio.wait_for(asyncio.shield(comfut), ack_timeout)
        except asyncio.CancelledError:
            raise ConnectionError

        state, msg, comnext = comfut.result()
        log.debug(f"Received ({state!r}, {msg!r})")

        if state == b"NEXT":
            if just_ack:
                # self.waiting_commands.append((commref, None))
                return b""
            else:
                assert comnext is not None
                await comnext
                state, msg, comnext2 = comnext.result()
                assert comnext2 is None
                log.debug(f"Received ({state!r}, {msg!r})")

        if state == b"OK":
            return msg
        elif state == b"ERRor":
            raise CommandError.parse(
                comm.decode(), commref.decode(), msg.decode().rstrip()
            ) from None
        else:  # pragma: no cover
            raise CommandError.parse(
                comm.decode(), commref.decode(), (state + b" " + msg).decode()
            )



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
# SPDX-License-Identifier: AGPL-3.0-only

from dataclasses import astuple

import hypothesis.strategies as st
import pytest  # noqa
from hypothesis import given

from qslib._util import _pp_seqsliceint
from qslib.qsconnection_async import FilterDataFilename


@given(
    stage=st.integers(0, 10000),
    cycle=st.integers(0, 10000),
    step=st.integers(0, 10000),
    point=st.integers(0, 10000),
    em=st.integers(1, 6),
    ex=st.integers(1, 6),
)
def test_fd_fn(stage, cycle, step, point, em, ex):
    fds = f"S{stage:02}_C{cycle:03}_T{step:02}" f"_P{point:04}_M{em}_X{ex}"
    fdfn = fds + "_filterdata.xml"

    x = FilterDataFilename.fromstring(fdfn)

    assert astuple(x) == ((ex, em, True), stage, cycle, step, point)

    assert fdfn == x.tostring()


def test_pp_seqsliceint():
    assert _pp_seqsliceint(5) == "5"
    assert _pp_seqsliceint(slice(1, 5)) == "1 to 5"
    assert _pp_seqsliceint(slice(5, None)) == "5 onwards"
    assert _pp_seqsliceint(slice(None, 5)) == "up to 5"
    assert _pp_seqsliceint([1, 2, 5]) == "[1, 2, 5]"
    assert _pp_seqsliceint(slice(1, 5, 2)) == "1 to 5 by step 2"
    with pytest.raises(TypeError):
        _pp_seqsliceint(5.5)  # type: ignore



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
# SPDX-License-Identifier: AGPL-3.0-only

import pytest

from qslib.scpi_commands import AccessLevel

levels = ["Guest", "Observer", "Controller", "Administrator", "Full"]

invalid = 5


def test_access():
    for l1 in levels:
        for l2 in levels:
            assert (AccessLevel(l1) < AccessLevel(l2)) == (
                levels.index(l1) < levels.index(l2)
            )
            assert (AccessLevel(l1) <= AccessLevel(l2)) == (
                levels.index(l1) <= levels.index(l2)
            )
            assert (AccessLevel(l1) > AccessLevel(l2)) == (
                levels.index(l1) > levels.index(l2)
            )
            assert (AccessLevel(l1) >= AccessLevel(l2)) == (
                levels.index(l1) >= levels.index(l2)
            )
            assert (AccessLevel(l1) == AccessLevel(l2)) == (
                levels.index(l1) == levels.index(l2)
            )
            assert (AccessLevel(l1) < l2) == (levels.index(l1) < levels.index(l2))
            assert (AccessLevel(l1) <= l2) == (levels.index(l1) <= levels.index(l2))
            assert (AccessLevel(l1) > l2) == (levels.index(l1) > levels.index(l2))
            assert (AccessLevel(l1) >= l2) == (levels.index(l1) >= levels.index(l2))
            assert (AccessLevel(l1) == l2) == (levels.index(l1) == levels.index(l2))
        with pytest.raises(ValueError):
            AccessLevel(l1) > invalid  # type: ignore
        with pytest.raises(ValueError):
            AccessLevel(l1) >= invalid  # type: ignore
        with pytest.raises(ValueError):
            AccessLevel(l1) < invalid  # type: ignore
        with pytest.raises(ValueError):
            AccessLevel(l1) <= invalid  # type: ignore
        with pytest.raises(ValueError):
            AccessLevel(l1) == invalid  # type: ignore
        assert str(AccessLevel(l1)) == l1



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
# SPDX-License-Identifier: AGPL-3.0-only

from pathlib import Path

import numpy as np
import pytest

from qslib import Experiment
from qslib.processors import (
    NormToMaxPerWell,
    NormToMeanPerWell,
    SmoothEMWMean,
    SmoothWindowMean,
    SubtractByMeanPerWell,
)


@pytest.fixture(scope="module")
def exp() -> Experiment:
    exp = Experiment.from_file("tests/test.eds")
    # We need better sample arrangements:
    exp.sample_wells["Sample 1"] = ["A7", "A8"]
    exp.sample_wells["Sample 2"] = ["A9", "A10"]
    exp.sample_wells["othersample"] = ["B7"]
    return exp


@pytest.fixture(scope="module")
def exp_reloaded(
    exp: Experiment, tmp_path_factory: pytest.TempPathFactory
) -> Experiment:
    tmp_path = tmp_path_factory.mktemp("exp")
    exp.save_file(tmp_path / "test_loaded.eds")
    return Experiment.from_file(tmp_path / "test_loaded.eds")


def test_props(exp: Experiment, exp_reloaded: Experiment) -> None:
    assert exp.name == "2020-02-20_170706"

    assert exp.info() == str(exp) == exp.summary()


def test_reload(exp: Experiment, exp_reloaded: Experiment) -> None:
    assert (exp.welldata == exp_reloaded.welldata).all().all()
    assert exp.name == exp_reloaded.name
    assert exp.protocol == exp_reloaded.protocol
    assert exp.plate_setup == exp_reloaded.plate_setup


def test_plot_ntmpw_smoothmw(exp: Experiment) -> None:
    axf, axt = exp.plot_over_time(
        process=[SmoothWindowMean(4), NormToMeanPerWell(1)], annotate_stage_lines=False
    )
    assert axf.get_ylabel() == "fluorescence (window mean 4, norm. to mean)"


def test_plot_emw_maxperwell(exp: Experiment) -> None:
    axf, axt = exp.plot_over_time(
        process=[SmoothEMWMean(alpha=0.1), NormToMaxPerWell(1)],
        annotate_stage_lines=False,
    )
    assert axf.get_ylabel() == "fluorescence (EMW-smoothed, norm. to max)"


def test_plot_subtrbymean(exp: Experiment) -> None:
    axf, axt = exp.plot_over_time(
        process=SubtractByMeanPerWell(1), annotate_stage_lines=False
    )
    assert axf.get_ylabel() == "fluorescence (subtr. by mean)"


def test_plots(exp: Experiment) -> None:
    axf, axt = exp.plot_over_time(
        legend=False, figure_kw={"constrained_layout": False}, annotate_stage_lines=True
    )

    # +2 here is for stage lines
    assert len(axf.get_lines()) == 5 * len(exp.all_filters) + 2
    assert np.allclose(axf.get_xlim(), (-0.004326652778519521, 0.09085970834891001))

    with pytest.raises(ValueError, match="Samples not found"):
        exp.plot_over_time("Sampl(e|a)")

    with pytest.raises(ValueError, match="Samples not found"):
        exp.plot_anneal_melt("Sampl(e|a)")

    import matplotlib.pyplot as plt

    _, ax = plt.subplots()
    axs = exp.plot_over_time(
        "Sample .*",
        "x1-m1",
        stages=2,
        temperatures=False,
        stage_lines=False,
        ax=ax,
        marker=".",
        legend=False,
    )

    axs2 = exp.plot_over_time(
        ["Sample 1", "Sample 2"],
        "x1-m1",
        stages=2,
        temperatures=False,
        stage_lines="fluorescence",
        annotate_stage_lines=("fluorescence", 0.1),
        marker=".",
        legend=True,
    )

    assert len(axs) == 1 == len(axs2)
    assert len(axs[0].get_lines()) == 4 == len(axs2[0].get_lines()) - 2

    axs = exp.plot_over_time("Sample .*")

    exp.plot_anneal_melt(samples="Sample 1")

    ax1 = exp.protocol.plot_protocol()
    ax2 = exp.plot_protocol()


def test_rawquant(exp: Experiment) -> None:
    exp.rawdata.loc[:, :]



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import annotations

import asyncio
import base64
import logging
import re
import zipfile
from asyncio.futures import Future
from contextlib import contextmanager
from dataclasses import dataclass
from functools import wraps
from typing import IO, TYPE_CHECKING, Any, Generator, Literal, cast, overload

import nest_asyncio

from qslib.qs_is_protocol import CommandError
from qslib.scpi_commands import AccessLevel, SCPICommand

from ._util import _unwrap_tags
from .protocol import Protocol
from .qsconnection_async import QSConnectionAsync

nest_asyncio.apply()

from .base import MachineStatus, RunStatus

log = logging.getLogger(__name__)

if TYPE_CHECKING:  # pragma: no cover
    import matplotlib.pyplot as plt

    from .experiment import Experiment


def _ensure_connection(level: AccessLevel = AccessLevel.Observer) -> Any:
    def wrap(func):
        @wraps(func)
        def wrapped(m: Machine, *args: Any, **kwargs: Any) -> Any:
            if m.automatic:
                with m.ensured_connection(level):
                    return func(m, *args, **kwargs)
            else:
                return func(m, *args, **kwargs)

        return wrapped

    return wrap


@dataclass(init=False)
class Machine:
    """
    A connection to a QuantStudio machine.  The connection can be opened and closed, and reused.
    A maximum access level can be set and changed, which will prevent the access level from going
    above that level.

    By default, the class tries to handle connections and access automatically.

    Parameters
    ----------

    host
        The host name or IP to connect to.

    password
        The password to use. Note that this class does not obscure or protect the password at all,
        because it should not be relied on for security.  See :ref:`access-and-security`  for more
        information.

    automatic
        Whether or not to automatically handle connection, disconnection, and where possible,
        access level.  Default True.

    max_access_level: "Observer", "Controller", "Administrator", or "Full"
        The maximum access level to allow.  This is *not* the initial access level, which
        will be Observer. The parameter can be changed later by changing the :code:`max_access_level`
        attribute.

    port
        The port to connect to. (Use the normal SCPI port, not the line-editor connection usually
        on 2323).  Default is 7000.

    """

    host: str
    password: str | None = None
    automatic: bool = True
    _max_access_level: AccessLevel = AccessLevel.Controller
    port: int = 7000
    _initial_access_level: AccessLevel = AccessLevel.Observer
    _current_access_level: AccessLevel = AccessLevel.Guest
    _connection: QSConnectionAsync | None = None

    def asdict(self, password: bool = False) -> dict[str, str | int]:
        d: dict[str, str | int] = {"host": self.host}
        if self.password and password:
            d["password"] = self.password
        if self.max_access_level != Machine._max_access_level:
            d["max_access_level"] = self.max_access_level.value
        if self.port != Machine.port:
            d["port"] = self.port
        if self.automatic != Machine.automatic:
            d["automatic"] = self.automatic

        return d

    @property
    def connection(self) -> QSConnectionAsync:
        """The :class:`QSConnectionAsync` for the connection, or a :class:`ConnectionError`."""
        if self._connection is None:
            raise ConnectionError
        else:
            return self._connection

    @connection.setter
    def connection(self, v: QSConnectionAsync | None) -> None:
        self._connection = v

    @property
    def max_access_level(self) -> AccessLevel:
        return self._max_access_level

    @max_access_level.setter
    def max_access_level(self, v: AccessLevel | str) -> None:
        if not isinstance(v, AccessLevel):
            self._max_access_level = AccessLevel(v)
        else:
            self._max_access_level = v

    def __init__(
        self,
        host: str,
        password: str | None = None,
        automatic: bool = True,
        max_access_level: AccessLevel | str = AccessLevel.Controller,
        port: int = 7000,
        _initial_access_level: AccessLevel | str = AccessLevel.Observer,
    ):
        self.host = host
        self.port = port
        self.password = password
        self.automatic = automatic
        self.max_access_level = AccessLevel(max_access_level)
        self._initial_access_level = AccessLevel(_initial_access_level)
        self._connection = None

    def connect(self) -> None:
        """Open the connection manually."""
        loop = asyncio.get_event_loop()

        self.connection = QSConnectionAsync(
            self.host,
            self.port,
            password=self.password,
            initial_access_level=self._initial_access_level,
        )
        loop.run_until_complete(self.connection.connect())
        self._current_access_level = self.get_access_level()[0]

    @property
    def connected(self) -> bool:
        """Whether or not there is a current connection to the machine.

        Note that when using automatic connections, this will usually be False,
        because connections will only be active when running a command.
        """
        if (not hasattr(self, "_connection")) or (self._connection is None):
            return False
        else:
            return self.connection.connected

    def __enter__(self) -> Machine:
        try:
            self.connect()
        except Exception as e:
            self.disconnect()
            raise e
        return self

    @_ensure_connection(AccessLevel.Guest)
    def run_command(self, command: str | SCPICommand) -> str:
        """Run a SCPI command, and return the response as a string.
        Waits for OK, not just NEXT.

        Parameters
        ----------
        command : str
            command to run

        Returns
        -------
        str
            Response message (after "OK", not including it)

        Raises
        ------
        CommandError
            Received an Error response.
        """
        if self.connection is None:
            raise ConnectionError(f"Not connected to {self.host}.")
        loop = asyncio.get_event_loop()
        try:
            return loop.run_until_complete(self.connection.run_command(command))
        except CommandError as e:
            e.__traceback__ = None
            raise e

    @_ensure_connection(AccessLevel.Guest)
    def run_command_to_ack(self, command: str | SCPICommand) -> str:
        """Run an SCPI command, and return the response as a string.
        Returns after the command is processed (OK or NEXT), but potentially
        before it has completed (NEXT).

        Parameters
        ----------
        commands
            command to run

        Returns
        -------
        str
            Response message (after "OK" or "NEXT", likely "" in latter case)

        Raises
        ------
        CommandError
            Received an Error response.
        """
        if self.connection is None:
            raise ConnectionError(f"Not connected to {self.host}")
        loop = asyncio.get_event_loop()
        try:
            return loop.run_until_complete(
                self.connection.run_command(command, just_ack=True)
            )
        except CommandError as e:
            e.__traceback__ = None
            raise e

    @_ensure_connection(AccessLevel.Guest)
    def run_command_bytes(self, command: str | bytes | SCPICommand) -> bytes:
        """Run an SCPI command, and return the response as bytes (undecoded).
        Returns after the command is processed (OK or NEXT), but potentially
        before it has completed (NEXT).

        Parameters
        ----------
        command
            command to run

        Returns
        -------
        bytes
            Response message (after "OK" or "NEXT", likely "" in latter case)

        Raises
        ------
        CommandError
            Received
        """
        if self.connection is None:
            raise ConnectionError(f"Not connected to {self.host}.")
        if isinstance(command, str):
            command = command.encode()
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.connection._protocol.run_command(command))

    @_ensure_connection(AccessLevel.Controller)
    def define_protocol(self, protocol: Protocol) -> None:
        """Send a protocol to the machine. This *is not related* to a particular
        experiment.  The name on the machine is set by the protocol.

        Parameters
        ----------
        protocol
            protocol to send
        """
        protocol.validate()
        self.run_command(protocol.to_scpicommand())

    @_ensure_connection(AccessLevel.Observer)
    def read_dir_as_zip(self, path: str, leaf: str = "FILE") -> zipfile.ZipFile:
        """Read a directory on the

        Parameters
        ----------
        path : str
            path on the machine
        leaf : str, optional
            leaf to use, by default "FILE"

        Returns
        -------
        zipfile.ZipFile
            the returned zip file
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self.connection.read_dir_as_zip(path, leaf))

    @overload
    def list_files(
        self,
        path: str,
        *,
        leaf: str = "FILE",
        verbose: Literal[True],
        recursive: bool = False,
    ) -> list[dict[str, Any]]:
        ...

    @overload
    def list_files(
        self,
        path: str,
        *,
        leaf: str = "FILE",
        verbose: Literal[False],
        recursive: bool = False,
    ) -> list[str]:
        ...

    @_ensure_connection(AccessLevel.Observer)
    def list_files(
        self,
        path: str,
        *,
        leaf: str = "FILE",
        verbose: bool = False,
        recursive: bool = False,
    ) -> list[str] | list[dict[str, Any]]:

        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self.connection.list_files(
                path, leaf=leaf, verbose=verbose, recursive=recursive
            )
        )

    @_ensure_connection(AccessLevel.Observer)
    def read_file(
        self, path: str, context: str | None = None, leaf: str = "FILE"
    ) -> bytes:
        """Read a file.

        Parameters
        ----------
        path : str
            File path on the machine.
        context : str | None (default None)
            Context.
        leaf: str (default FILE)

        Returns
        -------
        bytes
            returned file
        """
        return asyncio.get_event_loop().run_until_complete(
            self.connection.read_file(path, context, leaf)
        )

    @_ensure_connection(AccessLevel.Controller)
    def write_file(self, path: str, data: str | bytes) -> None:
        if isinstance(data, str):
            data = data.encode()

        self.run_command_bytes(
            b"FILE:WRITE "
            + path.encode()
            + b" <quote.base64>\n"
            + base64.encodebytes(data)
            + b"\n</quote.base64>"
        )

    @_ensure_connection(AccessLevel.Observer)
    def list_runs_in_storage(self) -> list[str]:
        """List runs in machine storage.

        Returns
        -------
        list[str]
            run filenames.  Retrieve with load_run_from_storage
            (to open as :any`Experiment`) or save_run_from_storage
            (to download and save it without opening.)
        """
        x = self.run_command("FILE:LIST? public_run_complete:")
        a = x.split("\n")[1:-1]
        return [
            re.sub("^public_run_complete:", "", s)[:-4] for s in a if s.endswith(".eds")
        ]

    @_ensure_connection(AccessLevel.Observer)
    def load_run_from_storage(self, path: str) -> "Experiment":  # type: ignore
        from .experiment import Experiment

        """Load a run from machine storage as an Experiment
        """
        return Experiment.from_machine_storage(self, path)

    @_ensure_connection(AccessLevel.Guest)
    def save_run_from_storage(
        self, machine_path: str, download_path: str | IO[bytes], overwrite: bool = False
    ) -> None:
        """Download a file from run storage on the machine.

        Parameters
        ----------
        machine_path : str
            filename on the machine
        download_path : str | IO[bytes]
            filename to download to, or an open file
        overwrite : bool, optional
            if False and provided a filename rather than an
            open file, will not overwrite existing filies; by default
            False
        """
        fdata = self.read_file(machine_path, context="public_run_complete")

        if not isinstance(download_path, str):
            file = download_path
            file.write(fdata)
        else:
            if overwrite:
                file = open(download_path, "wb")
            else:
                file = open(download_path, "xb")
            try:
                file.write(fdata)
            finally:
                file.close()

    @_ensure_connection(AccessLevel.Observer)
    def _get_log_from_byte(self, name: str | bytes, byte: int) -> bytes:
        logfuture: Future[
            tuple[bytes, bytes, Future[tuple[bytes, bytes, None]] | None]
        ] = asyncio.Future()
        if self.connection is None:
            raise Exception
        if isinstance(name, bytes):
            name = name.decode()
        self.connection._protocol.waiting_commands.append((b"logtransfer", logfuture))

        logcommand = self.connection._protocol.run_command(
            f"eval? session.writeQueue.put(('OK logtransfer \\<quote.base64\\>\\\\n'"
            f" + (lambda x: [x.seek({byte}), __import__('base64').encodestring(x.read())][1])"
            f"(open('/data/vendor/IS/experiments/{name}/apldbio/sds/messages.log')) +"
            " '\\</quote.base64\\>\\\\n', None))",
            ack_timeout=200,
        )

        loop = asyncio.get_event_loop()

        loop.run_until_complete(logcommand)

        loop.run_until_complete(logfuture)

        return base64.decodebytes(logfuture.result()[1][15:-17])

    @_ensure_connection(AccessLevel.Observer)
    def run_status(self) -> RunStatus:
        """Return information on the status of any run."""
        return RunStatus.from_machine(self)

    @_ensure_connection(AccessLevel.Observer)
    def machine_status(self) -> MachineStatus:
        """Return information on the status of the machine."""
        return MachineStatus.from_machine(self)

    @_ensure_connection(AccessLevel.Observer)
    def get_running_protocol(self) -> Protocol:
        p = _unwrap_tags(self.run_command("PROT? ${Protocol}"))
        pn, svs, rm = self.run_command(
            "RET ${Protocol} ${SampleVolume} ${RunMode}"
        ).split()
        p = f"PROT -volume={svs} -runmode={rm} {pn} " + p
        return Protocol.from_scpicommand(SCPICommand.from_string(p))

    def set_access_level(
        self,
        access_level: AccessLevel | str,
        exclusive: bool = False,
        stealth: bool = False,
    ) -> None:
        access_level = AccessLevel(access_level)

        if access_level > AccessLevel(self.max_access_level):
            raise ValueError(
                f"Access level {access_level} is above maximum {self.max_access_level}."
                " Change max_access level to continue."
            )

        self.run_command(
            f"ACC -stealth={stealth} -exclusive={exclusive} {access_level}"
        )
        log.debug(f"Took access level {access_level} {exclusive=} {stealth=}")
        self._current_access_level = access_level

    def get_access_level(
        self,
    ) -> tuple[AccessLevel, bool, bool]:
        ret = self.run_command("ACC?")
        m = re.match(r"^-stealth=(\w+) -exclusive=(\w+) (\w+)", ret)
        if m is None:
            raise ValueError(ret)
        level = AccessLevel(m[3])
        self._current_access_level = level
        return level, m[2] == "True", m[1] == "True"

    @property
    def access_level(self) -> AccessLevel:
        return self._current_access_level

    @access_level.setter
    def access_level(self, v: AccessLevel | str) -> None:
        with self.ensured_connection(AccessLevel.Guest):
            self.set_access_level(v)

    @_ensure_connection(AccessLevel.Controller)
    def drawer_open(self) -> None:
        """Open the machine drawer using the OPEN command. This will ensure proper
        cover/drawer operation.  It *will not check run status*, and will open and
        close the drawer during runs and potentially during imaging.
        """
        self.run_command("OPEN")

    @_ensure_connection(AccessLevel.Controller)
    def drawer_close(self, lower_cover: bool = True, check: bool = True) -> None:
        """Close the machine drawer using the OPEN command. This will ensure proper
        cover/drawer operation.  It *will not check run status*, and will open and
        close the drawer during runs and potentially during imaging.

        By default, it will lower the cover automaticaly after closing, use
        lower_cover=False to not do so.
        """
        self.run_command("CLOSE")
        if (drawerpos := self.drawer_position) != "Closed":
            log.error(f"Drawer position should be Closed, but is {drawerpos}.")
            if check:
                raise ValueError(f"Drawer position is {drawerpos}")
        if lower_cover:
            self.cover_lower(check=check, ensure_drawer=False)

    @property
    def status(self) -> RunStatus:
        """Return the current status of the run."""
        with self.ensured_connection(AccessLevel.Observer):
            return RunStatus.from_machine(self)

    @property
    def drawer_position(self) -> Literal["Open", "Closed", "Unknown"]:
        """Return the drawer position from the DRAW? command."""
        with self.ensured_connection(AccessLevel.Observer):
            d = self.run_command("DRAW?")
            if d not in ["Open", "Closed", "Unknown"]:
                raise ValueError(f"Drawer position {d} is not understood.")
            return cast(Literal["Open", "Closed", "Unknown"], d)

    @property
    def cover_position(self) -> Literal["Up", "Down", "Unknown", ""]:
        """Return the cover position from the ENG? command. Note that
        this does not always seem to work."""
        with self.ensured_connection(AccessLevel.Observer):
            f = self.run_command("ENG?")
            if f not in ["Up", "Down", "Unknown", ""]:
                raise ValueError(f"Cover position {f} is not understood.")
            if f == "":
                log.error("Cover position is blank. This should not happen.")
            return cast(Literal["Up", "Down", "Unknown", ""], f)

    @_ensure_connection(AccessLevel.Controller)
    def cover_lower(self, check: bool = True, ensure_drawer: bool = True) -> None:
        """Lower/engage the plate cover, closing the drawer if needed."""
        if ensure_drawer and (self.drawer_position in ("Open", "Unknown")):
            self.drawer_close(lower_cover=False, check=check)
        self.run_command("COVerDOWN")
        if (covpos := self.cover_position) != "Down":
            log.error(f"Cover position should be Down, but is {covpos}.")
            if check:
                raise ValueError(f"Cover position should be Down, but is {covpos}.")

    def __exit__(self, exc_type: type, exc: Exception, tb: Any) -> None:
        self.disconnect()

    def __del__(self) -> None:
        if self.connected:
            self.disconnect()

    def disconnect(self) -> None:
        """Cleanly disconnect from the machine."""
        if self.connection is None:
            raise ConnectionError(f"Not connected to {self.host}.")

        loop = asyncio.get_event_loop()

        loop.run_until_complete(self.connection.disconnect())
        self._connection = None
        self._current_access_level = AccessLevel.Guest

    @_ensure_connection(AccessLevel.Controller)
    def abort_current_run(self) -> None:
        """Abort (stop immediately) the current run."""
        self.run_command("AbortRun ${RunTitle}")

    @_ensure_connection(AccessLevel.Controller)
    def stop_current_run(self) -> None:
        """Stop (stop after cycle end) the current run."""
        self.run_command("StopRun ${RunTitle}")

    @_ensure_connection(AccessLevel.Controller)
    def pause_current_run(self) -> None:
        """Pause the current run now."""
        self.run_command_to_ack("PAUSe")

    @_ensure_connection(AccessLevel.Controller)
    def pause_current_run_at_temperature(self) -> None:
        raise NotImplementedError

    @_ensure_connection(AccessLevel.Controller)
    def resume_current_run(self) -> None:
        """Resume the current run."""
        self.run_command_to_ack("RESume")

    @property
    def power(self) -> bool:
        """Get and set the machine's operational power (lamp, etc) as a bool.

        Setting this to False will not turn off the machine, just power down
        the lamp, temperature control, etc.  It will do so even if there is
        currently a run.
        """
        with self.ensured_connection(AccessLevel.Observer):
            s = self.run_command("POW?").lower()
            if s == "on":
                return True
            elif s == "off":
                return False
            else:
                raise ValueError(f"Unexpected power status: {s}")

    @power.setter
    def power(self, value: Literal["on", "off", True, False]) -> None:
        with self.ensured_connection(AccessLevel.Controller):
            if value is True:
                value = "on"
            elif value is False:
                value = "off"
            self.run_command(f"POW {value}")

    @property
    def current_run_name(self) -> str | None:
        """Name of current run, or None if no run is active."""
        with self.ensured_connection(AccessLevel.Observer):
            out = self.run_command("RUNTitle?")
            if out == "-":
                return None
            else:
                return re.sub(r"(<([\w.]+)>)?([^<]+)(</[\w.]+>)?", r"\3", out)

    @_ensure_connection(AccessLevel.Controller)
    def restart_system(self) -> None:
        """Restart the system (both the InstrumentServer and android interface) by killing the zygote process."""
        self.run_command(SCPICommand("SYST:EXEC", "killall zygote"))

    @contextmanager
    def at_access(
        self,
        access_level: AccessLevel | str,
        exclusive: bool = False,
        stealth: bool = False,
    ) -> Generator[Machine, None, None]:
        fac, fex, fst = self.get_access_level()
        self.set_access_level(access_level, exclusive, stealth)
        log.debug(f"Took access level {access_level} {exclusive=} {stealth=}.")
        yield self
        self.set_access_level(fac, fex, fst)
        log.debug(
            f"Dropped access level {access_level}, returning to {fac} exclusive={fex} stealth={fst}."
        )

    @contextmanager
    def ensured_connection(
        self, access_level: AccessLevel = AccessLevel.Observer
    ) -> Generator[Machine, None, None]:
        if self.automatic:
            was_connected = self.connected
            if not was_connected:
                self.connect()
            old_access = self.access_level
            if old_access < access_level:
                self.set_access_level(access_level)
            yield self
            if not was_connected:
                self.disconnect()
            elif old_access < access_level:
                self.set_access_level(old_access)
        else:
            yield self



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

import pandas as pd


def _fdc_to_rawdata(fdc: pd.DataFrame, start_time: float) -> pd.DataFrame:
    ret: pd.DataFrame = fdc.loc[:, (slice(None), "fl")].copy()
    ret.columns = [f"{r}{c}" for r in "ABCDEFGH" for c in range(1, 13)]
    for i in range(0, 6):
        ret[f"temperature_{i+1}"] = fdc[f"A{(i+1)*2}", "rt"]
    ret["temperature_avg"] = ret.loc[:, slice("temperature_1", "temperature_6")].mean(
        axis=1
    )
    ret["timestamp"] = fdc["time", "timestamp"]
    ret["exptime"] = ret["timestamp"] - start_time
    ret["exphrs"] = (ret["timestamp"] - start_time) / 60 / 60
    # ret['time'] = ret['time'].astype('datetime64[s]') # fixme: should we?

    return ret



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import annotations

import os
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from glob import glob
from os import PathLike
from typing import List, Literal, Optional, Sequence, Union, cast

import numpy as np
import numpy.typing as npt
import pandas as pd

_UPPERS = "ABCDEFGHIJKLMNOP"


@dataclass(frozen=True, order=True, eq=True)
class FilterSet:
    """Representation of a filter set, potentially including the "quant"
    parameter used by HACFILT in SCPI protocols."""

    ex: int
    em: int
    quant: bool = True

    @classmethod
    def fromstring(cls, string: str | FilterSet | Sequence) -> FilterSet:
        if isinstance(string, FilterSet):
            return string
        # fixme: do validation
        if not isinstance(string, str):
            string = ",".join(string)

        if string.startswith("x"):
            return cls(int(string[1]), int(string[4]))
        elif string.startswith("M"):
            return cls(int(string[4]), int(string[1]))
        elif string.startswith("m"):
            return cls(int(string[4]), int(string[1]), string.endswith(",quant"))
        else:
            raise ValueError

    @property
    def lowerform(self) -> str:
        return f"x{self.ex}-m{self.em}"

    @property
    def upperform(self) -> str:
        return f"M{self.em}_X{self.ex}"

    @property
    def hacform(self) -> str:
        return f"m{self.em},x{self.ex}" + (",quant" if self.quant else "")

    def to_xml(self) -> ET.Element:
        e = ET.Element("CollectionCondition")
        ET.SubElement(e, "FilterSet", Emission=f"m{self.em}", Excitation=f"x{self.ex}")
        ET.SubElement(e, "Frames").text = "0"
        return e

    def __str__(self) -> str:
        return self.lowerform + ("-noquant" if not self.quant else "")


class FilterDataReading:
    stage: int
    step: int
    cycle: int
    point: int
    timestamp: float | None
    exposure: int
    filter_set: FilterSet
    well_fluorescence: npt.NDArray[np.float64]
    temperatures: npt.NDArray[np.float64]
    set_temperatures: npt.NDArray[np.float64] | None

    def __repr__(self):
        return (
            f"FilterDataReading(stage={self.stage}, cycle={self.cycle}, "
            f'step={self.step}, point={self.point}, filter_set="{self.filter_set}", '
            f"timestamp={self.timestamp}, ...)"
        )

    @classmethod
    def from_file(
        cls, path: str | PathLike[str], sds_dir: str | PathLike[str] | None
    ) -> FilterDataReading:
        p = ET.parse(path).find(".//PlateData")
        if p is None:
            raise ValueError(f"File {path} is not a valid filter data file.")
        return cls(p, sds_dir=sds_dir)

    def __init__(
        self,
        pde: ET.Element,
        timestamp: Optional[float] = None,
        sds_dir: Optional[str | PathLike[str]] = None,
        set_temperatures: Union[Literal["auto"], None, List[float]] = "auto",
    ):

        attribs = {
            cast(str, k.text).lower(): cast(str, v.text)
            for k, v in zip(
                pde.findall("Attribute/key"),
                pde.findall("Attribute/value"),
            )
        }

        self.temperatures = np.array(
            [float(x) for x in attribs["temperature"].split(",")], dtype=np.float64
        )

        self.stage = int(attribs["stage"])
        self.step = int(attribs["step"])
        self.cycle = int(attribs["cycle"])
        self.point = int(attribs["point"])
        self.exposure = int(attribs["exposure"])
        self.filter_set = FilterSet.fromstring(attribs["filter_set"])

        if set_temperatures == "auto":
            self.set_temperatures: Optional[np.ndarray] = cast(
                np.ndarray, self.temperatures
            ).round(2)
        elif set_temperatures is not None:
            assert len(set_temperatures) == len(self.temperatures)
            self.set_temperatures = np.array(set_temperatures)
        else:
            self.set_temperatures = None

        self.plate_rows = int(cast(str, cast(ET.Element, pde.find("Rows")).text))
        self.plate_cols = int(cast(str, cast(ET.Element, pde.find("Cols")).text))

        assert self.plate_cols % len(self.temperatures) == 0

        # todo: handle other cases
        assert self.plate_rows * self.plate_cols == 96
        assert len(self.temperatures) == 6

        self.well_fluorescence = cast(
            np.ndarray,
            np.fromstring(cast(ET.Element, pde.find("WellData")).text, sep=" "),
        )

        self.timestamp = timestamp

        if timestamp is None and sds_dir is not None:
            qs = glob(
                os.path.join(
                    sds_dir, "quant", self.filename_reading_string + "_E*.quant"
                )
            )
            qs.sort()
            self.set_timestamp_by_quantdata(open(qs[-1]).read())

    def set_timestamp_by_quantdata(self, qstring: str) -> float:
        qss = qstring.split("\n\n")[3].split("\n")
        assert len(qss) == 3
        assert qss[0] == "[conditions]"
        qd = {k: v for k, v in zip(qss[1].split("\t"), qss[2].split("\t"))}
        self.timestamp = float(qd["Time"])
        return self.timestamp

    @property
    def filename_reading_string(self) -> str:
        return (
            f"S{self.stage:02}_C{self.cycle:03}_T{self.step:02}_"
            f"P{self.point:04}_{self.filter_set.upperform}"
        )

    @property
    def well_temperatures(self) -> np.ndarray:
        return cast(
            np.ndarray,
            self.temperatures[
                np.tile(
                    np.arange(0, len(self.temperatures)).repeat(
                        self.plate_cols / len(self.temperatures)
                    ),
                    self.plate_rows,
                )
            ],
        )

    @property
    def well_set_temperatures(self) -> np.ndarray:
        if self.set_temperatures is not None:
            return cast(
                np.ndarray,
                self.set_temperatures[
                    np.tile(
                        np.arange(0, len(self.set_temperatures)).repeat(
                            self.plate_cols / len(self.set_temperatures)
                        ),
                        self.plate_rows,
                    )
                ],
            )
        else:
            return np.full(self.plate_cols * self.plate_rows, None)

    @property
    def plate_temperatures(self) -> np.ndarray:
        return self.well_temperatures.reshape(self.plate_rows, self.plate_cols)

    @property
    def plate_set_temperatures(self) -> np.ndarray:
        return self.well_set_temperatures.reshape(self.plate_rows, self.plate_cols)

    @property
    def plate_fluorescence(self) -> np.ndarray:
        return self.well_fluorescence.reshape(self.plate_rows, self.plate_cols)

    def to_lineprotocol(self, run_name: str = None, sample_array=None) -> List[str]:
        lines = []
        gs = f"filterdata,filter_set={self.filter_set}"
        assert self.timestamp
        es = " {}".format(int(self.timestamp * 1e9))

        wr = [
            (_UPPERS[rn], rn, c)
            for rn in range(0, self.plate_rows)
            for c in range(1, self.plate_cols + 1)
        ]

        for (r, rn, c), f, tr, ts in zip(
            wr,
            self.well_fluorescence,
            self.well_temperatures,
            self.well_set_temperatures,
        ):
            s = (
                f",row={r},col={c:02} fluorescence={f},temperature_read={tr}"
                f",stage={self.stage:02}i,cycle={self.cycle:03}i"
                f",step={self.step:02}i,point={self.point:04}i"
            )
            if sample_array is not None:
                s += f',sample="{sample_array[rn,c-1]}"'
            if run_name is not None:
                s += f',run_name="{run_name}"'
            if self.set_temperatures is not None:
                s += f",temperature_set={ts}"
            lines.append(gs + s + es)

        return lines


def df_from_readings(
    readings: List[FilterDataReading], start_time: float | None = None
) -> pd.DataFrame:
    f = [
        np.concatenate(
            (
                np.array([r.timestamp]),
                r.well_fluorescence,
                r.well_temperatures,
                r.well_set_temperatures,
                np.array([r.exposure]),
            )
        )
        for r in readings
    ]

    indices = pd.MultiIndex.from_tuples(
        [(r.filter_set.lowerform, r.stage, r.cycle, r.step, r.point) for r in readings],
        names=["filter_set", "stage", "cycle", "step", "point"],
    )

    # fr = readings[0]

    # wr = [
    #     f"{r}{c:02}"
    #     for r in _UPPERS[0 : fr.plate_rows]
    #     for c in range(1, fr.plate_cols + 1)
    # ]

    a = pd.DataFrame(
        f,
        index=indices,
        columns=pd.MultiIndex.from_tuples(
            [(cast(str, "time"), cast(str, "timestamp"))]
            + [
                (f"{r}{c}", v)
                for v in ["fl", "rt", "st"]
                for r in "ABCDEFGH"
                for c in range(1, 13)
            ]
            + [("exposure", "exposure")]
        ),
    )

    if start_time is not None:
        a[("time", "seconds")] = a[("time", "timestamp")] - start_time
        a[("time", "hours")] = a[("time", "seconds")] / 3600.0

    a.sort_index(inplace=True)

    return a.reindex(
        labels=pd.MultiIndex.from_tuples(
            [(cast(str, "time"), v) for v in ["seconds", "hours", "timestamp"]]
            + [
                (f"{r}{c}", v)
                for r in "ABCDEFGH"
                for c in range(1, 13)
                for v in ["fl", "rt", "st"]
            ]
            + [("exposure", "exposure")]
        ),
        axis="columns",  # type: ignore
    )



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

from __future__ import annotations

import time
from dataclasses import dataclass
from pathlib import Path

import click

from qslib.common import Experiment, Machine
from qslib.experiment import MachineError
from qslib.qs_is_protocol import (
    AccessLevelExceeded,
    AuthError,
    CommandError,
    InsufficientAccess,
)
from qslib.scpi_commands import AccessLevel


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "-o",
    "--output",
    help="output file name (defaults to experiment base filename + extension)",
    type=click.Path(),
)
@click.option(
    "-f",
    "--format",
    "format",
    help="Specify the output format (matplotlib formats are supported, eg, pdf, svg, png); defaults to pdf.",
    default="pdf",
)
@click.option(
    "-a",
    "--actual/--no-actual",
    default=False,
    help="NOT FINISHED Include actual temperature readings for finished experiments (not included by default).",
)
@click.option("-n", "--open/--no-open", help="Open the file after creating it.")
@click.argument("experiment", type=click.Path(exists=True))
def protocol_plot(experiment, output, format, actual, open) -> None:
    """Plot the temperature protocol in an experiment."""
    import matplotlib.pyplot as plt

    experiment = Path(experiment)

    exp = Experiment.from_file(experiment)

    if output is None:
        output = experiment.with_suffix("." + format)

    fig, ax = plt.subplots(figsize=(21.0 / 2.54, 15.0 / 2.54))

    exp.protocol.plot_protocol(ax)

    fig.savefig(output, format=format)

    if open:
        click.launch(str(output))


@cli.command()
@click.option(
    "-o",
    "--output",
    help="output file name (defaults to experiment base filename + extension)",
    type=click.Path(),
)
@click.option(
    "-n",
    "--open/--no-open",
    "openfile",
    help="Open the file after creating it.",
    default=True,
)
@click.argument("experiment", type=click.Path(exists=True))
def info_html(experiment, output, openfile) -> None:
    """Create an HTML summary of the experiment, and potentially open it."""
    import matplotlib.pyplot as plt

    experiment = Path(experiment)

    exp = Experiment.from_file(experiment)

    if output is None:
        output = experiment.with_suffix(".html")

    with open(output, "w") as f:
        f.write(exp.info_html())

    if openfile:
        click.launch(str(output))


@cli.command()
@click.argument("experiment", type=click.Path(exists=True))
def protocol_desc(experiment: str | Path) -> None:
    """Print a description of the protocol in an experiment."""
    experiment = Path(experiment)

    exp = Experiment.from_file(experiment)

    click.echo(str(exp.protocol))


@cli.command()
@click.argument("experiment", type=click.Path(exists=True))
def export_temperatures(experiment: str | Path) -> None:
    """Export temperature readings from an experiment, as a CSV file."""

    experiment = Path(experiment)

    exp = Experiment.from_file(experiment)

    assert exp.temperatures

    exp.temperatures.to_csv(path_or_buf=click.get_binary_stream("stdout"))  # type: ignore


@cli.command()
@click.argument("experiment", type=click.Path(exists=True))
def export_data(experiment: str | Path) -> None:
    """Export fluorescence reading data from an experiment, as a CSV file."""
    experiment = Path(experiment)

    exp = Experiment.from_file(experiment)

    exp.welldata.to_csv(path_or_buf=click.get_binary_stream("stdout"))  # type: ignore


@cli.command()
@click.argument("experiment", type=click.Path(exists=True))
def info(experiment: str) -> None:
    """Output information on the experiment."""

    exp = Experiment.from_file(experiment)

    click.echo(exp.info())


@cli.command()
@click.argument("experiment", type=click.Path(exists=True))
@click.argument("machine")
def run(experiment: str, machine: str) -> None:
    """Run an experiment."""
    exp = Experiment.from_file(experiment)

    import logging

    logging.basicConfig(level=logging.INFO)

    m = Machine(
        machine,
        max_access_level="Controller",
    )

    with m:
        exp.run(m)


@cli.command()
@click.argument("machine")
@click.argument("state", type=click.Choice(["on", "off"]))
def machine_power(machine: str, state: str) -> None:
    """Turn the lamp/heat-block on or off (if idle)."""
    m = Machine(
        machine,
        max_access_level="Controller",
    )

    with m:
        rs = m.run_status()
        ms = m.machine_status()
        mn = m.run_command("SYST:SETT:NICK?")

        if rs.state != "Idle":
            raise click.UsageError(
                f"Machine {mn} is currently running {rs.name}, not changing power state during run."
            )
        else:
            with m.at_access("Controller"):
                m.power = {"on": True, "off": False}[state]


@cli.command()
@click.argument("machine")
def machine_status(machine: str) -> None:
    """Print the current status of a machine."""
    m = Machine(
        machine,
        max_access_level="Observer",
    )

    m.connect()
    rs = m.run_status()
    ms = m.machine_status()
    mn = m.run_command("SYST:SETT:NICK?")
    m.disconnect()

    drawer = ms.drawer
    if drawer == "Closed":
        drawer = click.style(drawer, fg="green")
    if drawer == "Open":
        drawer = click.style(drawer, fg="yellow")
    if drawer == "Unknown":
        drawer = click.style(drawer, fg="red")

    cover = ms.cover
    if cover == "Down":
        cover = click.style(cover, fg="green")
    if cover == "Up":
        cover = click.style(cover, fg="blue")
    if cover == "Unknown":
        cover = click.style(cover, fg="red")

    lamp = ms.lamp_status.split()[0]
    if lamp == "Off":
        lamp = click.style(lamp, fg="yellow")
    if lamp == "On":
        lamp = click.style(lamp, fg="green")

    state = rs.state
    if state == "Running":
        state = click.style(state, fg="green")
    if state == "Idle":
        state = click.style(state, fg="blue")
    if state == "Error":
        state = click.style(state, fg="red")

    click.echo(f"Machine {mn} is {state}.")
    click.echo(
        f"Drawer is {drawer}, cover is {cover}, and lamp is {lamp} and {ms.led_temperature:.2f} °C."
    )

    click.echo(f"Cover temperature is {ms.cover_temperature:.2f} °C", nl=False)
    if ms.target_controlled["Cover"]:
        click.echo(
            f" and is controlled with a target of {ms.target_temperatures['Cover']:.2f} °C."
        )
    else:
        click.echo(f" and is uncontrolled.")

    click.echo(
        f"Block temperatures are "
        + ", ".join("{:.2f}".format(x) for x in ms.block_temperatures)
        + " °C."
    )
    click.echo(
        f"Sample temperatures are "
        + ", ".join("{:.2f}".format(x) for x in ms.block_temperatures)
        + " °C."
    )

    if all(ms.target_controlled[f"Zone{i}"] for i in range(1, 7)):
        click.echo(
            f"Zone temperatures are controlled with a target of "
            + ", ".join(
                "{:.2f}".format(ms.target_temperatures[f"Zone{i}"]) for i in range(1, 7)
            )
            + " °C."
        )
    else:
        click.echo(f"Zone temperatures are uncontrolled.")

    if rs.state != "Idle":
        click.echo(
            f"Run {rs.name}. Stage {rs.stage}/{rs.num_stages},"
            f" cycle {rs.cycle}/{rs.num_cycles}, and step {rs.step}."
        )

    del m


@cli.command()
@click.argument("machine")
def list_stored(machine: str) -> None:
    """List experiments stored on a machine."""
    m = Machine(
        machine,
        max_access_level="Observer",
    )

    with m:
        for f in m.list_runs_in_storage():
            click.echo(f)


@cli.command()
@click.argument("machine")
@click.argument("experiment")
@click.option("-o", "--output", type=click.Path())
def copy(
    machine: str,
    experiment: str,
    output: str | None,
) -> None:
    """Copy experiment from machine storage."""
    m = Machine(
        machine,
        max_access_level="Observer",
    )

    if output is None:
        output = experiment

    with m:
        exp = Experiment.from_machine(m, experiment)
        exp.save_file(output, update_files=False)


@dataclass
class OutP:
    level: bool

    def verbose(self, s: str):
        if self.level:
            click.secho(s, fg="blue")

    def out(self, s: str):
        click.secho(s, nl=False)

    def good(self, s: str):
        click.secho(s, fg="green")

    def warn(self, s: str):
        click.secho(s, fg="yellow")

    def error(self, s: str):
        click.secho(s, fg="red")


class NoNewAccess(BaseException):
    ...


class NoAccess(BaseException):
    ...


def set_default_access(m: Machine, p: OutP):
    limitcontexts = m.run_command("CONF* access accessLimits").split()

    p.verbose("Current access:")
    for lc in limitcontexts:
        al = m.run_command(f"CONF? access accessLimits {lc}")
        p.verbose(f"\t{lc}: {al}")

    p.out("Setting default (passwordless) maximum access level to CONTROLLER: ")
    m.run_command("CONF= access accessLimits default CONTROLLER")
    m.run_command("CONF= access accessLimits eth0 CONTROLLER")
    p.good("done.")


def add_controller_password(m: Machine, p: OutP, newpass: str):
    p.verbose("Current password information:")
    passwords = m.run_command("CONF* access secrets").split()
    for password in passwords:
        pp = m.run_command(f"CONF? access secrets {password}")
        p.verbose(f"\t{password} : {pp}")
    p.out("Adding controller password: ")
    m.run_command(f"CONF= -createMissing access secrets {newpass} CONTROLLER")
    p.good("done.")


def add_administrator_password(m: Machine, p: OutP, newpass: str):
    p.verbose("Current password information:")
    passwords = m.run_command("CONF* access secrets").split()
    for password in passwords:
        pp = m.run_command(f"CONF? access secrets {password}")
        p.verbose(f"\t{password} : {pp}")
    p.out("Adding administrator password: ")
    m.run_command(f"CONF= -createMissing access secrets {newpass} ADMINISTRATOR")
    p.good("done.")


def start_ssh_backup(m: Machine, p: OutP, sshpass: str):
    p.out("Starting SSH in case of failure: ")
    m.run_command(f"SYST:EXEC 'dropbear -Y appletini &'")
    psout = m.run_command("SYST:EXEC -verbose 'ps'")
    assert "dropbear" in psout
    p.good("done.")
    p.warn(
        f"If procedure fails, SSH is running with user root and password {sshpass}"
        " (but note options required to connect)."
    )


def check_access(
    host: str,
    p: OutP,
    controller_pw: str | None,
    admin_pw: str | None,
    default_controller: bool,
):
    errs: list[CommandError] = []
    if controller_pw is not None:
        p.out("Checking controller password (machine LED should turn yellow): ")
        m = Machine(
            host, max_access_level=AccessLevel.Controller, password=controller_pw
        )
        try:
            with m.ensured_connection(AccessLevel.Controller):
                m.run_command("LED:YELLOWON")
                time.sleep(2)
                m.run_command("LED:BLUEON")
            p.good("succeeded.")
        except AuthError as e:
            p.error("Controller password was not set successfully.")
            errs.append(e)

    if admin_pw is not None:
        p.out("Checking administrator password (machine LED should turn red): ")
        m = Machine(host, max_access_level=AccessLevel.Administrator, password=admin_pw)
        try:
            with m.ensured_connection(AccessLevel.Administrator):
                m.run_command("LED:REDON")
                time.sleep(2)
                m.run_command("LED:BLUEON")
            p.good("succeeded.")
        except AuthError as e:
            p.error("Administrator password was not set successfully.")
            errs.append(e)

    if default_controller:
        p.out(
            "Checking passwordless controller access (machine LED should turn green): "
        )
        m = Machine(host, max_access_level=AccessLevel.Controller)
        try:
            with m.ensured_connection(AccessLevel.Controller):
                m.run_command("LED:GREENON")
                time.sleep(2)
                m.run_command("LED:BLUEON")
            p.good("succeeded.")
        except AccessLevelExceeded as e:
            p.error("Setting passwordless controller access failed.")
            errs.append(e)

    if errs:
        raise NoNewAccess


def stop_ssh_backup(m: Machine, p: OutP):
    p.out("Stopping SSH: ")
    m.run_command(f"SYST:EXEC 'killall dropbear'")
    psout = m.run_command("SYST:EXEC -verbose 'ps'")
    assert "dropbear" not in psout
    p.good("done.")


def error_ssh(m: Machine, p: OutP):
    raise NotImplementedError


def restart_is(m: Machine, p: OutP):
    p.out("Restarting zygote on machine: ")
    with m:
        with m.at_access(AccessLevel.Controller):
            try:
                m.restart_system()
            except ConnectionError:
                p.good("succeeded.")
            else:
                p.good("command succeeded.")
    p.out("Waiting 30 seconds for restart: ")
    time.sleep(30)
    p.good("done.")
    p.out("Checking connection: ")
    with m:
        pass
    p.good("succeeded.")


@cli.command()
@click.argument("host")
@click.argument("current_password")
@click.option("-c", "--controller-password", default=None)
@click.option("-a", "--admin-password", default=None)
@click.option("-d", "--default-controller/--no-default-controller", default=False)
@click.option("-v", "--verbose/--no-verbose", default=False)
def setup_machine(
    host: str,
    current_password: str,
    controller_password: str | None,
    admin_password: str | None,
    default_controller: bool,
    verbose: bool,
):
    p = OutP(verbose)

    m = Machine(
        host, password=current_password, max_access_level=AccessLevel.Controller
    )

    try:
        with m.ensured_connection(AccessLevel.Controller):
            start_ssh_backup(m, p, "emergencypassword")
            if controller_password:
                add_controller_password(m, p, controller_password)
            if admin_password:
                add_administrator_password(m, p, admin_password)
            if default_controller:
                set_default_access(m, p)
    except AuthError as e:
        p.error(f"Current password was not valid.")
        if verbose:
            p.error(str(e))
    except InsufficientAccess as e:
        p.error(
            f"Current password was valid, but insufficient.  This script needs at least Controller access."
        )
        p.error(str(e))
    restart_is(m, p)

    try:
        check_access(host, p, controller_password, admin_password, default_controller)

        with m:
            with m.at_access(AccessLevel.Controller):
                stop_ssh_backup(m, p)
    except NoNewAccess:
        # Do we have *any* access?
        try:
            with Machine(
                host, password=current_password, max_access_level=AccessLevel.Controller
            ) as m:
                stop_ssh_backup(m, p)
        except Exception as e:
            p.error("Access failed even with old password.")
            p.error(f"Error was {e}")
            p.error("Machine can be accessed directly via SSH using:")
            p.error(
                f"    ssh root@{host} -p 2222 -o KexAlgorithms=diffie-hellman-group1-sha1"
                " -o HostKeyAlgorithms='+ssh-rsa'"
            )
            p.error(f"Password is 'emergencypassword'.")
            p.error("DO NOT REBOOT THE MACHINE or SSH access will be lost.")
        else:
            p.error("New access failed, but old password still works.")


if __name__ == "__main__":
    cli()



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
# SPDX-License-Identifier: AGPL-3.0-only

import asyncio
import re

import pytest
import pytest_asyncio

import qslib.qs_is_protocol
from qslib import Machine
from qslib.qsconnection_async import QSConnectionAsync


def crcb(crlist):
    async def _fakeserver_runner(sr: asyncio.StreamReader, sw: asyncio.StreamWriter):
        acc = "Guest"

        sw.write(
            b"READy -session=12345 -product=QuantStudio3_5 -version=1.3.0 -build=001 -capabilities=Index\n"
        )
        await sw.drain()

        while not sr.at_eof():
            line = await sr.readline()
            print(line)
            if x := re.match(rb"(\d+) ACC (\w+)", line, re.IGNORECASE):
                sw.write(b"OK " + x.group(1) + b"\n")
                acc = x.group(2).decode()
                continue

            if x := re.match(rb"(\d+) ACC?", line, re.IGNORECASE):
                sw.write(
                    b"OK "
                    + x.group(1)
                    + f" -stealth=False -exclusive=False {acc}".encode()
                    + b"\n"
                )
                continue

            if x := re.match(rb"(\d+) QUIT", line, re.IGNORECASE):
                sw.write(b"OK " + x.group(1) + b"\n")
                await sw.drain()
                sw.close()
                return

            if x := re.match(rb"(\d+) TESTKILLSERVER", line):
                sw.close()
                return

            if x := re.match(rb"(\d+) TESTERRORSERVER", line):
                sw.write(b"ERRor " + x.group(1) + b" testerror\n")
                await sw.drain()

            if x := re.match(rb"(\d+) TESTUNKNOWNRESPSERVER", line):
                sw.write(b"UNKNOWNCOMMAND " + x.group(1) + b" testerror\n")
                await sw.drain()
                sw.write(b"OK " + x.group(1) + b"\n")
                await sw.drain()

            if x := re.match(rb"(\d+) DOUBLEOK", line):
                sw.write(b"OK 12345 doubleok\n")
                await sw.drain()
                sw.write(b"OK " + x.group(1) + b"\n")
                await sw.drain()

            if x := re.match(rb"(\d+) TESTNEXTSERVER", line):
                sw.write(b"NEXT " + x.group(1) + b"\n")
                await sw.drain()
                sw.write(b"MESSage testservermessage ueao\n")
                await sw.drain()
                sw.write(b"MESSage testservermessage 123456789.021 ueao\n")
                await sw.drain()
                sw.write(b"OK " + x.group(1) + b"\n")
                await sw.drain()

            if x := re.match(rb"TESTNEXTSERVER", line):
                sw.write(b"NEXT " + line[:-1] + b"\n")
                await sw.drain()
                sw.write(b"MESSage testservermessage ueao\n")
                await sw.drain()
                sw.write(b"MESSage testservermessage 123456789.021 ueao\n")
                await sw.drain()
                sw.write(b"OK " + line[:-1] + b" return message" + b"\n")
                await sw.drain()

            if x := re.match(rb"(\d+) TESTNEXTSERVERDELAY", line):
                sw.write(b"NEXT " + x.group(1) + b"\n")
                await sw.drain()
                sw.write(b"MESSage testservermessage ueao\n")
                await asyncio.sleep(1)
                await sw.drain()
                sw.write(b"OK " + x.group(1) + b"\n")
                await sw.drain()

            for com, resp in crlist.items():
                if x := re.match(rb"(\d+) " + re.escape(com.encode()), line):
                    sw.write(b"OK " + x.group(1) + b" " + resp.encode() + b"\n")
                    await sw.drain()
                    break

    return _fakeserver_runner


@pytest.mark.asyncio
async def test_responses():
    srv = await asyncio.start_server(crcb({}), "localhost", 53533)

    async with srv:
        m = Machine("localhost", port=53533)

        m.connect()

        with pytest.raises(qslib.qs_is_protocol.CommandError):
            m.run_command("TESTERRORSERVER")

        m.run_command("TESTUNKNOWNRESPSERVER")
        # FIXME: this should check logs

        m.run_command_to_ack("TESTNEXTSERVER")

        m.run_command("TESTNEXTSERVER")

        m.run_command("TESTNEXTSERVERDELAY")

        m.disconnect()


@pytest.mark.asyncio
async def test_connection():
    srv = await asyncio.start_server(crcb({}), "localhost", 53533)

    async with srv:
        m = Machine("localhost", port=53533)

        assert m.connected is False

        m.connect()

        assert m.connected is True

        m.disconnect()

        assert m.connected is False

        m = Machine("localhost", port=53533)

        m.connect()

        assert m.connected is True

        assert m._connection is not None

        m._connection._protocol.waiting_commands.append((b"12345", None))

        with pytest.raises(ConnectionError):
            m.run_command_bytes(b"TESTKILLSERVER")

        assert m.connected is False


@pytest.mark.asyncio
async def test_acc_level_set():
    srv = await asyncio.start_server(crcb({}), "localhost", 53533)

    async with srv:
        m = Machine("localhost", port=53533)

        with m:
            with m.at_access("Observer"):
                assert m.get_access_level() == ("Observer", False, False)

            m.set_access_level("Observer")


@pytest.mark.asyncio
async def test_runtitle():
    srv = await asyncio.start_server(crcb({"RUNTitle?": "aoeu"}), "localhost", 53533)

    async with srv:
        with Machine("localhost", port=53533) as m:
            assert m.current_run_name == "aoeu"


@pytest.mark.asyncio
async def test_runtitle_not_running():
    srv = await asyncio.start_server(crcb({"RUNTitle?": "-"}), "localhost", 53533)

    async with srv:
        with Machine("localhost", port=53533) as m:
            assert m.current_run_name == None


@pytest.mark.asyncio
async def test_quote():
    msg = "<quote>a\nu\n\n \n <quote.2>C\n</quote.2>\n  </quote>"
    srv = await asyncio.start_server(crcb({"TESTQUOTE": msg}), "localhost", 53533)

    async with srv:
        with Machine("localhost", port=53533) as m:
            assert m.run_command("TESTQUOTE") == msg


@pytest.mark.asyncio
async def test_invalid_quote():
    msg = "<quote>a\nu\n</quote.2>\n  </quote>"
    srv = await asyncio.start_server(crcb({"TESTQUOTE": msg}), "localhost", 53533)

    async with srv:
        with Machine("localhost", port=53533) as m:
            with pytest.raises(ConnectionError):
                m.run_command("TESTQUOTE")


@pytest.mark.asyncio
async def test_nonuid_nonreturn():
    srv = await asyncio.start_server(crcb({}), "localhost", 53533)

    async with srv:
        qsa = QSConnectionAsync("localhost", 53533)

        assert qsa.connected is False

        async with qsa:
            assert b"return message\n" == await qsa._protocol.run_command(
                "TESTNEXTSERVER", uid=False, just_ack=False
            )

            qsa._protocol.waiting_commands.append((b"12345", None))
            await qsa._protocol.run_command("DOUBLEOK")



# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
#
# SPDX-License-Identifier: AGPL-3.0-only

"""Code for handling plate setup."""
from __future__ import annotations

import xml.etree.ElementTree as ET
from dataclasses import dataclass, field
from io import BytesIO
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from uuid import uuid1

import attrs
import numpy as np
import pandas as pd
import tabulate

from .qsconnection_async import QSConnectionAsync

_WELLNAMES = [x + str(y) for x in "ABCDEFGH" for y in range(1, 13)]

_WELLNAMESET = set(_WELLNAMES)

_WELLALPHREF = [(x, f"{y}") for x in "ABCDEFGH" for y in range(1, 13)]


def _process_color_from_str_int(x: str) -> Tuple[int, int, int, int]:
    """From a string that represents a signed int32 (this choice make no sense),
    interpret it as a unsigned 32-bit integer, then unpack the bits to get R,G,B,A."""

    color_bytes: Tuple[int, int, int, int] = tuple(
        int(b) for b in int(x).to_bytes(4, "little", signed=True)
    )  # type: ignore

    return color_bytes


def _color_to_str_int(x: Tuple[int, int, int, int]) -> str:
    return str(int.from_bytes(bytes(x), "little", signed=True))


def _str_or_list_to_list(v: str | Sequence[str]) -> list[str]:
    if isinstance(v, str):
        return [v]
    return list(v)


@attrs.define(init=False)
class Sample:
    name: str
    color: Tuple[int, int, int, int] = attrs.field(default=(255, 0, 0, 255))
    properties: dict[str, str] = attrs.field(factory=dict)
    description: str | None = None
    wells: list[str] = attrs.field(
        factory=list, converter=_str_or_list_to_list, on_setattr=attrs.setters.convert
    )

    def __init__(
        self,
        name: str,
        uuid: str | None = None,
        color: Tuple[int, int, int, int] = (255, 0, 0, 255),
        properties: dict[str, str] | None = None,
        description: str | None = None,
        wells: str | list[str] | None = None,
    ) -> None:
        if properties is None:
            properties = dict()
        if not "SP_UUID" in properties:
            if not uuid:
                uuid = uuid1().hex
            properties["SP_UUID"] = uuid
        if wells is None:
            wells = list()
        self.__attrs_init__(name, color, properties, description, wells=wells)  # type: ignore

    @property
    def uuid(self) -> str:
        return self.properties["SP_UUID"]

    @uuid.setter
    def uuid(self, val: str) -> None:
        self.properties["SP_UUID"] = val

    @classmethod
    def from_platesetup_sample(cls, se: ET.Element) -> Sample:
        name = se.findtext("Name") or "Unnamed"
        color = _process_color_from_str_int(se.findtext("Color") or "-1")

        keys = [
            x.text for x in se.findall("CustomProperty/Property") if x.text is not None
        ]
        values = [
            x.text for x in se.findall("CustomProperty/Value") if x.text is not None
        ]
        properties = {key: value for key, value in zip(keys, values)}

        return cls(
            name=name,
            color=color,
            properties=properties,
            description=se.findtext("Description"),
        )

    def to_xml(self) -> ET.Element:
        x = ET.Element("Sample")
        ET.SubElement(x, "Name").text = self.name
        ET.SubElement(x, "Color").text = _color_to_str_int(self.color)
        if self.description:
            ET.SubElement(x, "Description").text = self.description
        u = ET.SubElement(x, "CustomProperty")
        for key in self.properties.keys():
            ET.SubElement(u, "Property").text = key
        ET.SubElement(u, "Property").text = "SP_UUID"
        for value in self.properties.values():
            ET.SubElement(u, "Value").text = value
        ET.SubElement(u, "Value").text = self.uuid
        return x


@attrs.define()
class _SampleWellsView(Mapping[str, list[str]]):
    samples_by_name: Dict[str, Sample]

    def __getitem__(self, name: str) -> list[str]:
        return self.samples_by_name[name].wells

    def __setitem__(self, name: str, wells: str | list[str]) -> None:
        if isinstance(wells, str):
            wells = [wells]
        try:
            self.samples_by_name[name].wells = wells
        except KeyError:
            self.samples_by_name[name] = Sample(name=name, wells=wells)

    def __len__(self) -> int:
        return len(self.samples_by_name)

    def __iter__(self) -> Iterator[str]:
        return iter(self.samples_by_name)


@dataclass
class PlateSetup:
    samples_by_name: Dict[str, Sample]

    @property
    def sample_wells(self):
        return _SampleWellsView(self.samples_by_name)

    @classmethod
    def from_platesetup_xml(cls, platexml: ET.Element) -> PlateSetup:  # type: ignore
        # assert platexml.find("PlateKind/Type").text == "TYPE_8X12"

        sample_fvs = platexml.findall(
            "FeatureMap/Feature/Id[.='sample']/../../FeatureValue"
        )

        samples_by_name: Dict[str, Sample] = dict()
        samples_by_uuid: Dict[str, Sample] = dict()

        sample_wells: Dict[str, list[str]] = dict()

        for fv in sample_fvs:
            if x := fv.findtext("Index"):
                idx = int(x)
            else:
                raise ValueError
            if y := fv.find("FeatureItem/Sample"):
                sample = Sample.from_platesetup_sample(y)
            if sample.name in samples_by_name.keys():
                assert sample == samples_by_name[sample.name]
                assert sample == samples_by_uuid[sample.uuid]
                sample_wells[sample.name].append(_WELLNAMES[idx])
            else:
                assert sample.uuid not in samples_by_uuid.keys()
                samples_by_name[sample.name] = sample
                samples_by_uuid[sample.uuid] = sample
                sample_wells[sample.name] = [_WELLNAMES[idx]]

        return cls(sample_wells, samples_by_name)

    def __init__(
        self,
        sample_wells: Mapping[str, str | List[str]] | None = None,
        samples: Iterable[Sample] | Mapping[str, Sample] = tuple(),
    ) -> None:
        if isinstance(samples, Mapping):
            self.samples_by_name = dict(samples)
        else:
            self.samples_by_name = {s.name: s for s in samples}

        if sample_wells is not None:
            for name, wells in sample_wells.items():
                if name in self.samples_by_name:
                    if isinstance(wells, str):
                        wells = [wells]
                    self.samples_by_name[name].wells = wells
                else:
                    self.samples_by_name[name] = Sample(name, wells=wells)

    @property
    def well_sample(self):
        well_sample_name = pd.Series(np.full(8 * 12, None, object), index=_WELLNAMES)
        for s, ws in self.sample_wells.items():
            for w in ws:
                well_sample_name.loc[w] = s
        return well_sample_name

    def get_wells(self, samples_or_wells: str | Sequence[str]) -> list[str]:
        """
        Given a sample, well, or list of the two, returns the corresponding
        wells.  Note that this relies on samples not having well-like names.
        """
        wells = []

        if isinstance(samples_or_wells, str):
            samples_or_wells = [samples_or_wells]

        for sw in samples_or_wells:
            if sw.upper() in _WELLNAMESET:
                wells.append(sw.upper())
            else:
                wells += self.sample_wells[sw]

        return wells

    def get_descriptive_string(self, name: str) -> str:
        if (w := name.upper()) in _WELLNAMESET:
            return w
        sample = self.samples_by_name[name]
        return sample.description or sample.name

    def well_samples_as_array(self) -> np.ndarray[Any, Any]:
        return self.well_sample.to_numpy().reshape((8, 12))

    def to_lineprotocol(self, timestamp: int, run_name: str | None = None) -> list[str]:
        if run_name:
            rts = f',run_name="{run_name}"'
        else:
            rts = ""
        return [
            f'platesetup,row={r},col={c} sample="{s}"{rts} {timestamp}'
            for ((r, c), s) in zip(_WELLALPHREF, self.well_sample)
        ]

    @classmethod
    def from_array(
        cls, array: Union[np.ndarray, pd.DataFrame], *, make_unique: bool = False
    ) -> PlateSetup:
        raise NotImplemented

    @classmethod
    def from_tsv(cls, tsvstr: str) -> PlateSetup:
        raise NotImplemented

    def to_table(
        self,
        headers: Sequence[Union[str, int]] = list(range(1, 13)),
        tablefmt: str = "orgtbl",
        showindex: Sequence[str] = tuple("ABCDEFGH"),
        **kwargs: Any,
    ) -> str:
        return tabulate.tabulate(
            self.well_samples_as_array(),
            tablefmt=tablefmt,
            headers=[str(x) for x in headers],
            showindex=showindex,
            **kwargs,
        )

    def update_xml(self, root: ET.Element) -> None:
        samplemap = root.find("FeatureMap/Feature/Id[.='sample']/../..")
        e: Optional[ET.Element]
        if not samplemap:
            e = ET.SubElement(root, "FeatureMap")
            v = ET.SubElement(e, "Feature")
            ET.SubElement(v, "Id").text = "sample"
            ET.SubElement(v, "Name").text = "sample"
            samplemap = e
        ws = np.array(self.well_sample)
        for welli in range(0, 96):
            if ws[welli]:
                e = samplemap.find(f"FeatureValue/Index[.='{welli}']/../FeatureItem")
                if not e:
                    e = ET.SubElement(samplemap, "FeatureValue")
                    ET.SubElement(e, "Index").text = str(welli)
                    e = ET.SubElement(e, "FeatureItem")
                if s := e.find("Sample"):
                    e.remove(s)
                e.append(self.samples_by_name[ws[welli]].to_xml())
            else:
                if e := samplemap.find(f"FeatureValue/Index[.='{welli}']/.."):
                    samplemap.remove(e)

    @classmethod
    async def from_machine(
        cls, c: QSConnectionAsync, runtitle: Optional[str] = None
    ) -> PlateSetup:
        s = await c.get_sds_file("plate_setup.xml", runtitle=runtitle)
        x = ET.parse(BytesIO(s), parser=None)
        return cls.from_platesetup_xml(x.getroot())

    def __repr__(self) -> str:
        return f"PlateSetup(samples {self.sample_wells.keys()}))"

    def __str__(self) -> str:
        s = ""
        if self.sample_wells:
            s += "Plate setup:\n\n"
            for sample, wells in self.sample_wells.items():
                s += f" - {sample}: {wells}\n"
        return s

    def _repr_markdown_(self) -> str:
        if len(self.sample_wells) < 12:
            return str(self)
        else:
            return self.to_table(tablefmt="pipe")
